from mayan.apps.dependencies.classes import (
    JavaScriptDependency, PythonDependency
)

JavaScriptDependency(
    module=__name__, name='signature_pad', version_string='=4.0.4'
)
PythonDependency(
    module=__name__, name='CairoSVG', version_string='==2.7.1'
)



from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_cache_partition_purged, event_cache_purged
from ..permissions import (
    permission_cache_purge, permission_cache_view
)

from .mixins import CacheTestMixin, CacheViewTestMixin


class CacheViewTestCase(
    CacheTestMixin, CacheViewTestMixin, GenericViewTestCase
):
    def test_cache_detail_view_no_permission(self):
        self._create_test_cache()

        self._clear_events()

        response = self._request_test_cache_detail_view()
        self.assertNotContains(
            response=response, text=self._test_cache.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_detail_view_with_access(self):
        self._create_test_cache()

        self.grant_access(
            obj=self._test_cache, permission=permission_cache_view
        )

        self._clear_events()

        response = self._request_test_cache_detail_view()
        self.assertContains(
            response=response, text=self._test_cache.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_list_view_with_no_permission(self):
        self._create_test_cache()

        self._clear_events()

        response = self._request_test_cache_list_view()
        self.assertNotContains(
            response=response, text=self._test_cache.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_list_view_with_access(self):
        self._create_test_cache()

        self.grant_access(
            obj=self._test_cache, permission=permission_cache_view
        )

        self._clear_events()

        response = self._request_test_cache_list_view()
        self.assertContains(
            response=response, text=self._test_cache.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_purge_view_no_permission(self):
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

        cache_total_size = self._test_cache.get_total_size()

        self._clear_events()

        response = self._request_test_cache_purge_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(cache_total_size, self._test_cache.get_total_size())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_purge_view_with_access(self):
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

        self.grant_access(
            obj=self._test_cache, permission=permission_cache_purge
        )

        cache_total_size = self._test_cache.get_total_size()

        self._clear_events()

        response = self._request_test_cache_purge_view()
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(
            cache_total_size, self._test_cache.get_total_size()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_cache)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_cache_partition)
        self.assertEqual(events[0].verb, event_cache_partition_purged.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_cache)
        self.assertEqual(events[1].verb, event_cache_purged.id)

    def test_cache_multiple_purge_view_no_permission(self):
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

        cache_total_size = self._test_cache.get_total_size()

        self._clear_events()

        response = self._request_test_cache_multiple_purge_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(cache_total_size, self._test_cache.get_total_size())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cache_multiple_purge_view_with_access(self):
        self._create_test_cache()
        self._create_test_cache_partition()
        self._create_test_cache_partition_file()

        self.grant_access(
            obj=self._test_cache, permission=permission_cache_purge
        )

        cache_total_size = self._test_cache.get_total_size()

        self._clear_events()

        response = self._request_test_cache_multiple_purge_view()
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(
            cache_total_size, self._test_cache.get_total_size()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_cache)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_cache_partition)
        self.assertEqual(events[0].verb, event_cache_partition_purged.id)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_cache)
        self.assertEqual(events[1].verb, event_cache_purged.id)



from contextlib import contextmanager
import imaplib
import logging

from django.utils.encoding import force_bytes, force_str
from django.utils.translation import ugettext_lazy as _

from mayan.apps.sources.exceptions import SourceException
from mayan.apps.sources.source_backends.base import SourceBackend

from .literals import (
    DEFAULT_EMAIL_IMAP_MAILBOX, DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
    DEFAULT_EMAIL_IMAP_STORE_COMMANDS
)
from .mixins import SourceBackendMixinEmail

logger = logging.getLogger(name=__name__)


class SourceBackendIMAPEmail(SourceBackendMixinEmail, SourceBackend):
    label = _('IMAP email')

    @classmethod
    def get_form_field_widgets(cls):
        widgets = super().get_form_field_widgets()

        widgets.update(
            {
                'search_criteria': {
                    'class': 'django.forms.widgets.Textarea'
                },
                'store_commands': {
                    'class': 'django.forms.widgets.Textarea'
                }
            }
        )

        return widgets

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'mailbox': {
                    'class': 'django.forms.fields.CharField',
                    'default': DEFAULT_EMAIL_IMAP_MAILBOX,
                    'help_text': _(
                        'IMAP Mailbox from which to check for messages.'
                    ),
                    'kwargs': {
                        'max_length': 64,
                    },
                    'label': _('Mailbox')
                },
                'search_criteria': {
                    'blank': True,
                    'class': 'django.forms.fields.CharField',
                    'default': DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
                    'help_text': _(
                        'Criteria to use when searching for messages to '
                        'process. Use the format specified in '
                        'https://tools.ietf.org/html/rfc2060.html#section-6.4.4'
                    ),
                    'label': _('Search criteria'),
                    'null': True,
                },
                'store_commands': {
                    'blank': True,
                    'class': 'django.forms.fields.CharField',
                    'default': DEFAULT_EMAIL_IMAP_STORE_COMMANDS,
                    'help_text': _(
                        'IMAP STORE command to execute on messages after '
                        'they are processed. One command per line. Use '
                        'the commands specified in '
                        'https://tools.ietf.org/html/rfc2060.html#section-6.4.6 '
                        'or the custom commands for your IMAP server.'
                    ),
                    'label': _('Store commands'),
                    'null': True, 'required': False
                },
                'execute_expunge': {
                    'class': 'django.forms.fields.BooleanField',
                    'default': True,
                    'help_text': _(
                        'Execute the IMAP expunge command after processing '
                        'each email message.'
                    ),
                    'label': _('Execute expunge'),
                    'required': False
                },
                'mailbox_destination': {
                    'blank': True,
                    'class': 'django.forms.fields.CharField',
                    'help_text': _(
                        'IMAP Mailbox to which processed messages will '
                        'be copied.'
                    ),
                    'label': _('Destination mailbox'),
                    'max_length': 96,
                    'null': True,
                    'required': False
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _('IMAP protocol'), {
                    'fields': (
                        'mailbox', 'search_criteria', 'store_commands',
                        'execute_expunge', 'mailbox_destination'
                    )
                }
            ),
        )

        return fieldsets

    @contextmanager
    def _get_server(self):
        logger.debug(msg='Starting IMAP email connection')
        logger.debug(
            'host: %s', self.kwargs['host']
        )
        logger.debug(
            'ssl: %s', self.kwargs['ssl']
        )

        if self.kwargs['ssl']:
            imap_module_name = 'IMAP4_SSL'
        else:
            imap_module_name = 'IMAP4'

        imap_module = getattr(imaplib, imap_module_name)

        kwargs = {
            'host': self.kwargs['host'], 'port': self.kwargs['port']
        }

        credential = self.get_credential()
        password = credential.get('password')
        username = credential.get('username')

        with imap_module(**kwargs) as server:
            server.login(password=password, user=username)

            try:
                server.select(
                    mailbox=self.kwargs['mailbox']
                )
            except Exception as exception:
                raise SourceException(
                    'Error selecting mailbox: {}; {}'.format(
                        self.kwargs['mailbox'], exception
                    )
                )
            else:
                yield server

    def action_file_delete(self, message_id):
        with self._get_server() as server:
            if self.kwargs['store_commands']:
                for command in self.kwargs['store_commands'].split('\n'):
                    try:
                        args = [message_id]
                        args.extend(
                            command.strip().split(' ')
                        )
                        server.uid('STORE', *args)
                    except Exception as exception:
                        raise SourceException(
                            'Error executing IMAP store command "{}" '
                            'on message uid {}; {}'.format(
                                command, message_id, exception
                            )
                        )

            if self.kwargs['mailbox_destination']:
                try:
                    server.uid(
                        'COPY', message_id, self.kwargs['mailbox_destination']
                    )
                except Exception as exception:
                    raise SourceException(
                        'Error copying message uid {} to mailbox {}; '
                        '{}'.format(
                            message_id, self.kwargs['mailbox_destination'], exception
                        )
                    )

            if self.kwargs['execute_expunge']:
                server.expunge()

    def action_file_get(self, message_id):
        message_id = force_bytes(s=message_id)

        with self._get_server() as server:
            status, data = server.uid('FETCH', message_id, '(RFC822)')

            try:
                yield from self.process_message(
                    message=data[0][1]
                )
            except Exception as exception:
                raise SourceException(
                    'Error processing message uid: {}; {}'.format(
                        message_id, exception
                    )
                )

    def get_stored_file_list(self):
        with self._get_server() as server:
            try:
                search_criteria = self.kwargs['search_criteria'].strip().split()

                status, data = server.uid(
                    'SEARCH', None, *search_criteria
                )
            except Exception as exception:
                raise SourceException(
                    'Error executing search command; {}'.format(exception)
                )
            else:
                if data:
                    # data is a space separated sequence of message uids.
                    uids = data[0].split()

                    logger.debug(
                        'messages count: %s', len(uids)
                    )
                    logger.debug('message uids: %s', uids)

                    for uid in uids:
                        logger.debug('message uid: %s', uid)

                        # uids are bytes. Convert to unicode to allow
                        # serialization for the background task.
                        yield force_str(s=uid)



from django.contrib import messages
from django.db import models
from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.generics import FormView, SingleObjectListView
from mayan.apps.views.view_mixins import ExternalContentTypeObjectViewMixin

from ..classes import EventType, ModelEventType
from ..forms import (
    EventTypeUserRelationshipFormSet, ObjectEventTypeUserRelationshipFormSet
)
from ..icons import (
    icon_event_types_subscriptions_list,
    icon_object_event_type_user_subscription_list,
    icon_user_object_subscriptions_list
)
from ..models import ObjectEventSubscription, StoredEventType
from ..permissions import permission_events_view


class EventTypeSubscriptionListView(FormView):
    form_class = EventTypeUserRelationshipFormSet
    main_model = 'user'
    submodel = StoredEventType
    view_icon = icon_event_types_subscriptions_list

    def dispatch(self, *args, **kwargs):
        EventType.refresh()
        return super().dispatch(*args, **kwargs)

    def form_valid(self, form):
        try:
            for instance in form:
                instance.save()
        except Exception as exception:
            messages.error(
                message=_(
                    'Error updating event subscription; %s'
                ) % exception, request=self.request
            )
        else:
            messages.success(
                message=_('Event subscriptions updated successfully'),
                request=self.request
            )

        return super().form_valid(form=form)

    def get_extra_context(self):
        return {
            'form_display_mode_table': True,
            'object': self.get_object(),
            'title': _(
                'Event subscriptions'
            ) % self.get_object()
        }

    def get_initial(self):
        obj = self.get_object()
        initial = []

        for element in self.get_queryset():
            initial.append(
                {
                    'main_model': self.main_model,
                    'stored_event_type': element,
                    'user': obj
                }
            )
        return initial

    def get_object(self):
        return self.request.user

    def get_queryset(self):
        # Return the queryset by name from the sorted list of the class
        event_type_ids = [event_type.id for event_type in EventType.all()]

        # Preserve the queryset order to that of the sorted ID list by
        # namespace label and event label.
        # Create a conditional statement to annotate each row with the sort
        # index number. Then sort the query set by the custom sort index
        # field.
        when_list = []
        for sort_index, event_type_id in enumerate(iterable=event_type_ids):
            when_list.append(
                models.When(name=event_type_id, then=sort_index)
            )

        queryset = self.submodel.objects.filter(name__in=event_type_ids)
        queryset = queryset.annotate(
            sort_index=models.Case(
                *when_list, output_field=models.IntegerField()
            )
        )
        return queryset.order_by('sort_index')


class ObjectEventTypeSubscriptionListView(
    ExternalContentTypeObjectViewMixin, FormView
):
    external_object_permission = permission_events_view
    form_class = ObjectEventTypeUserRelationshipFormSet
    view_icon = icon_object_event_type_user_subscription_list

    def dispatch(self, *args, **kwargs):
        EventType.refresh()
        return super().dispatch(*args, **kwargs)

    def form_valid(self, form):
        try:
            for instance in form:
                instance.save()
        except Exception as exception:
            messages.error(
                message=_(
                    'Error updating object event subscription; %s'
                ) % exception, request=self.request
            )
        else:
            messages.success(
                message=_(
                    'Object event subscriptions updated successfully'
                ), request=self.request
            )

        return super().form_valid(form=form)

    def get_extra_context(self):
        return {
            'form_display_mode_table': True,
            'object': self.external_object,
            'title': _(
                'Event subscriptions for: %s'
            ) % self.external_object
        }

    def get_initial(self):
        initial = []

        for element in self.get_queryset():
            initial.append(
                {
                    'object': self.external_object,
                    'stored_event_type': element,
                    'user': self.request.user
                }
            )
        return initial

    def get_queryset(self):
        return ModelEventType.get_for_instance(instance=self.external_object)


class UserObjectSubscriptionList(SingleObjectListView):
    object_permission = permission_events_view
    view_icon = icon_user_object_subscriptions_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_user_object_subscriptions_list,
            'no_results_text': _(
                'Subscribe to the events of an object to received '
                'notifications when those events occur.'
            ),
            'no_results_title': _('There are no object event subscriptions'),
            'object': self.request.user,
            'title': _('Object event subscriptions')
        }

    def get_source_queryset(self):
        return ObjectEventSubscription.objects.filter(user=self.request.user)



from django.utils.translation import ugettext_lazy as _


class CommentBusinessLogicMixin:
    def get_user_label(self):
        if self.user.get_full_name():
            return self.user.get_full_name()
        else:
            return self.user.username
    get_user_label.short_description = _('User')



from mayan.apps.django_gpg.permissions import permission_key_sign
from mayan.apps.django_gpg.tests.mixins import KeyTestMixin
from mayan.apps.documents.events import (
    event_document_file_created, event_document_file_edited,
    event_document_version_created, event_document_version_page_created
)
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.literals import TEST_FILE_SMALL_PATH

from ..events import event_embedded_signature_created
from ..permissions import permission_document_file_sign_embedded

from .mixins import EmbeddedSignatureViewTestMixin


class EmbeddedSignaturesViewTestCase(
    KeyTestMixin, EmbeddedSignatureViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_embedded_signature_create_view_with_no_permission(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self._clear_events()

        response = self._request_test_document_file_embedded_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_embedded_signature_create_view_with_document_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_embedded
        )

        self._clear_events()

        response = self._request_test_document_file_embedded_signature_create_view()
        self.assertEqual(response.status_code, 200)

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_embedded_signature_create_view_with_key_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._clear_events()

        response = self._request_test_document_file_embedded_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_embedded_signature_create_view_with_full_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_embedded
        )
        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._clear_events()

        response = self._request_test_document_file_embedded_signature_create_view()
        self.assertEqual(response.status_code, 302)

        self._test_document.refresh_from_db()

        self.assertTrue(
            str(self._test_document.file_latest.pk) in response.url
        )

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document_file = self._test_document.file_latest
        test_document_version = self._test_document.versions.last()

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(
            events[3].target, test_document_version.pages.first()
        )
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

        self.assertEqual(
            events[4].action_object,
            self._test_document.file_latest.signatures.first().embeddedsignature
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, self._test_document_file)
        self.assertEqual(events[4].verb, event_embedded_signature_created.id)

    def test_trashed_document_embedded_signature_create_view_with_full_access(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_embedded
        )
        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_embedded_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self._test_document.refresh_from_db()

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signature_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging
from pathlib import Path

from django.apps import apps
from django.contrib.auth import get_user_model
from django.db import OperationalError

from mayan.celery import app

from .document_file_tasks import task_document_file_create
from .utils import execute_callback

logger = logging.getLogger(name=__name__)


@app.task(bind=True, ignore_results=True, retry_backoff=True)
def task_document_upload(
    self, document_type_id, shared_uploaded_file_id,
    callback_dict=None, description=None, label=None, language=None,
    user_id=None
):
    callback_dict = callback_dict or {}

    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )
    SharedUploadedFile = apps.get_model(
        app_label='storage', model_name='SharedUploadedFile'
    )
    User = get_user_model()

    try:
        document_type = DocumentType.objects.get(pk=document_type_id)
        shared_uploaded_file = SharedUploadedFile.objects.get(
            pk=shared_uploaded_file_id
        )
        if user_id:
            user = User.objects.get(pk=user_id)
        else:
            user = None
    except OperationalError as exception:
        raise self.retry(exc=exception)

    description = description or ''

    label = label or Path(
        str(shared_uploaded_file)
    ).name

    try:
        document = document_type.documents_create(
            description=description, label=label, language=language,
            user=user
        )
    except OperationalError as exception:
        logger.error(
            'Operational error creating new document of type: %s, '
            'label: %s; %s. Retrying.', document_type, label, exception
        )
        raise self.retry(exc=exception)
    except Exception as exception:
        logger.critical(
            'Unexpected exception while creating new document of type: %s, '
            'label: %s; %s', document_type, label, exception
        )
        raise
    else:
        execute_callback(
            callback_dict=callback_dict, document=document,
            name='post_document_create'
        )

        task_document_file_create.apply_async(
            kwargs={
                'callback_dict': callback_dict,
                'document_id': document.pk,
                'filename': label,
                'is_document_upload_sequence': True,
                'shared_uploaded_file_id': shared_uploaded_file_id,
                'user_id': user_id
            }
        )



import re

from dateutil.parser import parse

from django.utils.translation import ugettext_lazy as _

from .classes import MetadataParser


class DateAndTimeParser(MetadataParser):
    label = _('Date and time parser')

    def execute(self, input_data):
        return parse(input_data).isoformat()


class DateParser(MetadataParser):
    label = _('Date parser')

    def execute(self, input_data):
        return parse(input_data).date().isoformat()


class RegularExpressionParser(MetadataParser):
    arguments = ('pattern', 'replacement')
    label = _('Regular expression parser')

    def execute(self, input_data):
        return re.sub(
            pattern=self.kwargs['pattern'], repl=self.kwargs['replacement'],
            string=input_data
        )


class TimeParser(MetadataParser):
    label = _('Time parser')

    def execute(self, input_data):
        return parse(input_data).time().isoformat()



from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.document_indexing.models.index_instance_models import IndexInstanceNode
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin

from .literals import TEST_DOCUMENT_FILE_CONTENT_INDEX_NODE_TEMPLATE
from .mixins import DocumentFileContentTestMixin


class DocumentFileContentIndexingTestCase(
    DocumentFileContentTestMixin, IndexTemplateTestMixin,
    GenericDocumentTestCase
):
    _test_index_template_node_expression = TEST_DOCUMENT_FILE_CONTENT_INDEX_NODE_TEMPLATE
    auto_upload_test_document = True

    def test_indexing_document_file_not_parsed(self):
        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document
            ).exists()
        )

    def test_indexing_document_file_parsed_delete(self):
        self._create_test_document_file_parsed_content()
        value = ' '.join(self._test_document.file_latest.content())

        self._do_test_document_file_parsed_content_delete()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )

    def test_indexing_document_file_parsed_finished(self):
        self._create_test_document_file_parsed_content()
        value = ' '.join(self._test_document.file_latest.content())

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )



import re

from dateutil.parser import parse

from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from .classes import MetadataValidator


class DateAndTimeValidator(MetadataValidator):
    label = _('Date and time validator')

    def execute(self, input_data):
        parse(input_data).isoformat()


class DateValidator(MetadataValidator):
    label = _('Date validator')

    def execute(self, input_data):
        parse(input_data).date().isoformat()


class RegularExpressionValidator(MetadataValidator):
    arguments = ('pattern',)
    label = _('Regular expression validator')

    def execute(self, input_data):
        result = re.fullmatch(
            pattern=self.kwargs['pattern'], string=input_data
        )
        if not result:
            raise ValidationError(
                message=_('The input string does not match the pattern.')
            )


class TimeValidator(MetadataValidator):
    label = _('Time validator')

    def execute(self, input_data):
        parse(input_data).time().isoformat()



from django.conf import settings
from django.contrib import messages
from django.http import Http404
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.generics import FormView, SingleObjectListView

from .classes import SettingNamespace, Setting
from .forms import SettingForm
from .icons import (
    icon_setting_namespace_detail, icon_setting_namespace_list,
    icon_setting_edit
)
from .permissions import permission_settings_edit, permission_settings_view


class SettingEditView(FormView):
    form_class = SettingForm
    view_icon = icon_setting_edit
    view_permission = permission_settings_edit

    def dispatch(self, request, *args, **kwargs):
        if settings.COMMON_DISABLE_LOCAL_STORAGE:
            messages.warning(
                message=_(
                    'Local storage is currently disabled, changes to '
                    'setting values will not be saved or take effect.'
                ), request=self.request
            )

        self.object = self.get_object()

        return super().dispatch(request=request, *args, **kwargs)

    def form_valid(self, form):
        self.object.value = form.cleaned_data['value']
        Setting.save_configuration()
        messages.success(
            message=_('Setting updated successfully.'),
            request=self.request
        )
        return super().form_valid(form=form)

    def get_extra_context(self):
        return {
            'hide_link': True,
            'navigation_object_list': ('object', 'setting_namespace'),
            'object': self.object,
            'setting_namespace': self.object.namespace,
            'title': _('Edit setting: %s') % self.object
        }

    def get_initial(self):
        return {'setting': self.object}

    def get_object(self):
        return Setting.get(
            global_name=self.kwargs['setting_global_name']
        )

    def get_post_action_redirect(self):
        return reverse(
            viewname='settings:setting_namespace_detail', kwargs={
                'namespace_name': self.object.namespace.name
            }
        )


class SettingNamespaceDetailView(SingleObjectListView):
    view_icon = icon_setting_namespace_detail
    view_permission = permission_settings_view

    def get_extra_context(self):
        return {
            'hide_object': True,
            'object': self.get_namespace(),
            'subtitle': _(
                'Settings inherited from an environment variable take '
                'precedence and cannot be changed in this view. '
            ),
            'title': _('Settings in namespace: %s') % self.get_namespace()
        }

    def get_namespace(self):
        try:
            return SettingNamespace.get(
                name=self.kwargs['namespace_name']
            )
        except KeyError:
            raise Http404(
                _('Namespace: %s, not found') % self.kwargs['namespace_name']
            )

    def get_source_queryset(self):
        return self.get_namespace().settings


class SettingNamespaceListView(SingleObjectListView):
    extra_context = {
        'hide_link': True,
        'title': _('Setting namespaces')
    }
    view_icon = icon_setting_namespace_list
    view_permission = permission_settings_view

    def get_source_queryset(self):
        return SettingNamespace.get_all()



from rest_framework import status
from rest_framework.response import Response

from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import classproperty
from django.utils.translation import ugettext_lazy as _

from ..exceptions import SourceActionExceptionInterfaceArgumentMissing

from .arguments import argument_request
from .interface_arguments import SourceBackendActionInterfaceArgument


class SourceBackendActionInterfaceMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )

        mcs.update_arguments(new_class=new_class)

        return new_class

    @classmethod
    def update_arguments(mcs, new_class):
        argument_list = []
        for mro_klass in new_class.mro():
            argument_base = getattr(mro_klass, 'Argument', None)
            if argument_base:
                for item_name in dir(argument_base):
                    item = getattr(argument_base, item_name)
                    try:
                        if isinstance(item, SourceBackendActionInterfaceArgument):
                            item.name = item_name

                            if item.required and item.has_default:
                                raise ImproperlyConfigured(
                                    'Argument `{}` cannot be required and '
                                    'also provide a default.'.format(
                                        item.name
                                    )
                                )
                            elif not item.required and not item.has_default:
                                raise ImproperlyConfigured(
                                    'Optional argument `{}` must specify a '
                                    'default.'.format(
                                        item.name
                                    )
                                )

                            if item not in argument_list:
                                argument_list.append(item)
                    except TypeError:
                        """Expected error. Ignore and continue."""

        setattr(new_class, 'arguments', argument_list)


class SourceBackendActionInterface(
    metaclass=SourceBackendActionInterfaceMetaclass
):
    hidden = False

    @classproperty
    def arguments_visible(cls):
        for argument in cls.arguments:
            if not argument.hidden:
                yield argument

    @classproperty
    def name(cls):
        return str(cls.__name__)

    def __init__(self, action):
        """
        Load:     Data arrives in the context, is processed by the interface,
                  and stored in the action_kwargs.

        Retrieve: Action data is stored in action_data, context is used as
                  persistent storage for the MRO and the final return value
                  is set in interface_result.
        """
        self.action = action
        self.action_data = None
        self.action_kwargs = {}
        self.context = {}
        self.interface_result = None

    def get_confirmation_context(self, request):
        return {}

    def load(self, **kwargs):
        for argument in self.arguments:
            try:
                value = kwargs.pop(argument.name)
            except KeyError:
                if argument.required:
                    raise SourceActionExceptionInterfaceArgumentMissing(
                        'Argument `{}` missing for interface `{}`.'.format(
                            argument.name, self.name
                        )
                    )
                else:
                    value = argument.default

            self.context[argument.name] = value

        self.process_interface_context()

        return self.action_kwargs

    def process_action_data(self):
        return

    def process_interface_context(self):
        return

    def retrieve(self, data, context=None):
        self.context = context
        self.action_data = data
        self.process_action_data()
        return self.interface_result


class SourceBackendActionInterfaceRequest(SourceBackendActionInterface):
    class Argument:
        request = argument_request


class SourceBackendActionInterfaceRequestRESTAPI(
    SourceBackendActionInterfaceRequest
):
    def process_action_data(self):
        if self.action_data:
            headers = self.context['view'].get_success_headers(
                data=self.action_data
            )

            self.interface_result = Response(
                data=self.action_data, headers=headers,
                status=status.HTTP_200_OK
            )
        else:
            self.interface_result = Response(status=status.HTTP_200_OK)


class SourceBackendActionInterfaceRequestView(
    SourceBackendActionInterfaceRequest
):
    """
    Interface for HTML views.
    """


class SourceBackendActionInterfaceRequestViewForm(
    SourceBackendActionInterfaceRequestView
):
    class Argument:
        forms = SourceBackendActionInterfaceArgument(
            help_text=_(
                'Forms containing the upload data generated by the '
                'upload views.'
            )
        )


class SourceBackendActionInterfaceTask(SourceBackendActionInterface):
    """
    Interface that every action needs to received the background task
    arguments.
    """



from .literals import TEST_USER_LOCALE_LANGUAGE, TEST_USER_LOCALE_TIMEZONE


class UserLocaleProfileViewMixin:
    def _request_test_current_user_locale_profile_detail_view(self):
        return self._request_test_user_locale_profile_detail_view(
            user=self._test_case_user
        )

    def _request_test_current_user_locale_profile_edit_view(
        self, follow=None
    ):
        return self._request_test_user_locale_profile_edit_view(
            user=self._test_case_user, follow=follow
        )

    def _request_test_super_user_locale_profile_detail_view(self):
        return self._request_test_user_locale_profile_detail_view(
            user=self._test_super_user
        )

    def _request_test_super_user_locale_profile_edit_view(self):
        return self._request_test_user_locale_profile_edit_view(
            user=self._test_super_user
        )

    def _request_test_user_locale_profile_detail_view(self, user=None):
        user = user or self._test_user

        return self.get(
            viewname='locales:user_locale_profile_detail', kwargs={
                'user_id': user.pk
            }
        )

    def _request_test_user_locale_profile_edit_view(
        self, follow=None, user=None
    ):
        user = user or self._test_user

        return self.post(
            viewname='locales:user_locale_profile_edit', kwargs={
                'user_id': user.pk
            }, data={
                'language': TEST_USER_LOCALE_LANGUAGE,
                'timezone': TEST_USER_LOCALE_TIMEZONE
            }, follow=follow
        )



from django.utils.translation import ugettext_lazy as _

from .models import UserMailer
from .permissions import permission_user_mailer_use


class ObjectEmailActionMixin:
    form_field_widgets = {
        'body': {
            'class': 'django.forms.widgets.Textarea', 'kwargs': {}
        }
    }
    form_fields = {
        'recipient': {
            'label': _('Recipient'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    'Email address of the recipient. Can be multiple '
                    'addresses separated by comma or semicolon. A template '
                    'can be used to reference properties of the document.'
                ),
                'required': True
            }
        },
        'cc': {
            'label': _('CC'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    'Address used in the "Bcc" header when sending the '
                    'email. Can be multiple addresses '
                    'separated by comma or semicolon. A template can be '
                    'used to reference properties of the document.'
                ),
                'required': False
            }
        },
        'bcc': {
            'label': _('BCC'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    'Address used in the "Bcc" header when sending the '
                    'email. Can be multiple addresses '
                    'separated by comma or semicolon. A template can be '
                    'used to reference properties of the document.'
                ),
                'required': False
            }
        },
        'reply_to': {
            'label': _('Reply to'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    'Address used in the "Reply-To" header when sending '
                    'the email. Can be multiple addresses separated by '
                    'comma or semicolon. A template can be used to '
                    'reference properties of the document.'
                ),
                'required': False
            }
        },
        'subject': {
            'label': _('Subject'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    'Subject of the email. Can be a string or a template.'
                ),
                'required': False
            }
        },
        'body': {
            'label': _('Body'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    'Body of the email to send. Can be a string or '
                    'a template.'
                ),
                'required': False
            }
        },
        'attachment': {
            'label': _('Attachment'),
            'class': 'django.forms.BooleanField', 'default': False,
            'help_text': _(
                'Attach an object to the email.'
            ),
            'required': False
        }
    }
    label = _('Send object via email')
    permission = permission_user_mailer_use

    @classmethod
    def get_form_fields(cls, *args, **kwargs):
        fields = super().get_form_fields(*args, **kwargs)

        fields.update(
            {
                'mailing_profile': {
                    'class': 'mayan.apps.views.fields.FormFieldFilteredModelChoice',
                    'help_text': _(
                        'Mailing profile to use when sending the email.'
                    ),
                    'kwargs': {
                        'source_queryset': UserMailer.objects.filter(enabled=True),
                        'permission': cls.permission
                    },
                    'label': _('Mailing profile'),
                    'required': True
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _('Mailing profile'), {
                    'fields': ('mailing_profile',)
                }
            ), (
                _('Parties'), {
                    'fields': ('recipient', 'cc', 'bcc', 'reply_to')
                },
            ), (
                _('Content'), {
                    'fields': ('subject', 'attachment', 'body')
                },
            )
        )
        return fieldsets

    def execute(self, context):
        user_mailer = self.get_user_mailer()
        user_mailer.send_object(
            **self.get_execute_data(context=context)
        )

    def get_execute_data(self, context):
        recipient = self.render_field(
            field_name='recipient', context=context
        )
        cc = self.render_field(
            field_name='cc', context=context
        )
        bcc = self.render_field(
            field_name='bcc', context=context
        )
        reply_to = self.render_field(
            field_name='reply_to', context=context
        )
        subject = self.render_field(
            field_name='subject', context=context
        )
        body = self.render_field(
            field_name='body', context=context
        )

        kwargs = {
            'bcc': bcc, 'body': body, 'cc': cc,
            'obj': self.get_object(context=context), 'reply_to': reply_to,
            'subject': subject, 'to': recipient
        }

        return kwargs

    def get_object(self, context):
        return NotImplementedError

    def get_user_mailer(self):
        return UserMailer.objects.get(
            pk=self.kwargs['mailing_profile']
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.dependencies.classes import BinaryDependency

from .backends.tesseract import Tesseract

tesseract = Tesseract(auto_initialize=False)
tesseract.read_settings()

BinaryDependency(
    legal_text='''
        The code in this repository is licensed under the Apache License,
        Version 2.0 (the "License");
        you may not use this file except in compliance with the License.
        You may obtain a copy of the License at

           http://www.apache.org/licenses/LICENSE-2.0

        Unless required by applicable law or agreed to in writing, software
        distributed under the License is distributed on an "AS IS" BASIS,
        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
        implied. See the License for the specific language governing
        permissions and limitations under the License.
    ''', help_text=_('Free Open Source OCR Engine'), label='Tesseract',
    module=__name__, name='tesseract', path=tesseract.tesseract_binary_path
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_MIME_TYPE_BACKEND, DEFAULT_MIME_TYPE_BACKEND_ARGUMENTS
)

namespace = SettingNamespace(
    label=_('MIME types'), name='mime_types'
)

setting_backend = namespace.add_setting(
    default=DEFAULT_MIME_TYPE_BACKEND, global_name='MIME_TYPE_BACKEND',
    help_text=_(
        'Path to the class to use when to detect file MIME types.'
    )
)
setting_backend_arguments = namespace.add_setting(
    default=DEFAULT_MIME_TYPE_BACKEND_ARGUMENTS,
    global_name='MIME_TYPE_BACKEND_ARGUMENTS', help_text=_(
        'Arguments to pass to the MIME_TYPE_BACKEND.'
    )
)



from ..models.recently_accessed_document_models import RecentlyAccessedDocument

from .base import GenericDocumentTestCase


class RecentlyAccessedDocumentModelTestCase(GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_trashed_document_accessed_document_add(self):
        self._create_test_document_stub()
        self._test_document.delete()
        self._test_document.add_as_recent_document_for_user(
            user=self._test_case_user
        )

        self.assertFalse(
            RecentlyAccessedDocument.valid.filter(
                document=self._test_document
            ).exists()
        )



from django.db import migrations

from ..compatibility import RGBColorField

COLOR_BLUE = 'blu'
COLOR_CORAL = 'crl'
COLOR_CYAN = 'cya'
COLOR_GREENYELLOW = 'gry'
COLOR_KHAKI = 'kki'
COLOR_LIGHTGREY = 'lig'
COLOR_MAGENTA = 'mag'
COLOR_ORANGE = 'org'
COLOR_RED = 'red'
COLOR_YELLOW = 'yel'

RGB_VALUES = {
    COLOR_BLUE: '#0000ff',
    COLOR_CORAL: '#ff7f50',
    COLOR_CYAN: '#00ffff',
    COLOR_GREENYELLOW: '#adff2f',
    COLOR_KHAKI: '#f0e68c',
    COLOR_LIGHTGREY: '#d3d3d3',
    COLOR_MAGENTA: '#ff00ff',
    COLOR_ORANGE: '#ffa500',
    COLOR_RED: '#ff0000',
    COLOR_YELLOW: '#ffff00',
}


def code_convert_color_names_to_rgb(apps, schema_editor):
    Tag = apps.get_model(app_label='tags', model_name='Tag')

    for tag in Tag.objects.using(alias=schema_editor.connection.alias).all():
        tag.selection = RGB_VALUES[tag.color]
        tag.save()


class Migration(migrations.Migration):
    dependencies = [
        ('tags', '0001_initial')
    ]

    operations = [
        migrations.AddField(
            model_name='tag',
            name='selection',
            field=RGBColorField(default='#FFFFFF'),
            preserve_default=False,
        ),
        migrations.RunPython(code=code_convert_color_names_to_rgb)
    ]



from django import forms as django_forms
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string


class FormMixinDynamicFields:
    def __init__(self, schema, *args, **kwargs):
        self.schema = schema

        super().__init__(*args, **kwargs)

        self.fieldsets = self.schema.get(
            'fieldsets', ()
        )
        self.fieldset_exclude_list = self.schema.get(
            'fieldset_exclude_list', ()
        )

        widgets = self.schema.get(
            'widgets', {}
        )

        field_order = self.get_field_order()

        for field_name in field_order:
            field_data = self.schema['fields'][field_name]
            field_class = import_string(
                dotted_path=field_data['class']
            )
            kwargs = {
                'label': field_data['label'],
                'required': field_data.get('required', True),
                'initial': field_data.get('default', None),
                'help_text': field_data.get('help_text')
            }

            widget = widgets.get(field_name)
            if widget:
                kwargs['widget'] = import_string(
                    dotted_path=widget['class']
                )(
                    **widget.get(
                        'kwargs', {}
                    )
                )

            kwargs.update(
                field_data.get(
                    'kwargs', {}
                )
            )
            self.fields[field_name] = field_class(**kwargs)

    def get_field_order(self):
        return self.schema.get(
            'field_order', self.schema['fields'].keys()
        )

    @property
    def media(self):
        """
        Append the media of the dynamic fields to the normal fields' media.
        """
        media = super().media
        media += django_forms.Media(
            **self.schema.get(
                'media', {}
            )
        )
        return media


class FormMixinFieldsets:
    fieldset_exclude_list = None
    fieldsets = None

    def get_fieldset_exclude_list(self):
        return self.fieldset_exclude_list or ()

    def get_fieldsets(self):
        if self.fieldsets:
            fieldsets_field_list = []
            for fieldset, data in self.fieldsets:
                fieldsets_field_list.extend(
                    data['fields']
                )

            set_fields = set(self.fields)
            set_fieldsets = set(fieldsets_field_list)

            fieldset_exclude_list = self.get_fieldset_exclude_list()

            if fieldset_exclude_list:
                set_fields -= set(fieldset_exclude_list)
                set_fieldsets -= set(fieldset_exclude_list)

            if set_fields != set_fieldsets:
                raise ImproperlyConfigured(
                    'Mismatch fieldset fields: {fields} in form `{form}`'.format(
                        fields=', '.join(
                            set_fields.symmetric_difference(set_fieldsets)
                        ), form=self.__class__.__name__
                    )
                )

            return self.fieldsets
        else:
            return (
                (
                    None, {
                        'fields': tuple(self.fields)
                    }
                ),
            )



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('storage', '0007_auto_20210218_0708')
    ]

    operations = [
        migrations.RemoveField(
            model_name='downloadfile',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='downloadfile',
            name='object_id',
        ),
        migrations.RemoveField(
            model_name='downloadfile',
            name='permission',
        ),
        migrations.AddField(
            model_name='downloadfile',
            name='user',
            field=models.ForeignKey(
                default=1, editable=False,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='download_files', to='auth.user',
                verbose_name='User'
            ),
            preserve_default=False,
        )
    ]



from django.db import models


class RGBColorField(models.CharField):
    """Compatibility subclass for colorful.fields.RGBColorField."""
    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 7
        super().__init__(*args, **kwargs)



from mayan.apps.common.tests.mixins import ManagementCommandTestMixin
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.storages import storage_document_files
from mayan.apps.mime_types.tests.mixins import MIMETypeBackendMixin

from ..literals import COMMAND_NAME_STORAGE_PROCESS

from .mixins import StorageProcessorTestMixin


class StorageProcessManagementCommandTestCase(
    MIMETypeBackendMixin, ManagementCommandTestMixin,
    StorageProcessorTestMixin, GenericDocumentTestCase
):
    _test_management_command_name = COMMAND_NAME_STORAGE_PROCESS

    def _get_test_management_command_keyword_arguments(self, reverse=None):
        keyword_arguments = {
            'app_label': 'documents',
            'defined_storage_name': storage_document_files.name,
            'log_file': str(self.path_test_file),
            'model_name': 'DocumentFile',
            'reverse': reverse
        }
        return keyword_arguments

    def _upload_and_call(self):
        self.defined_storage.dotted_path = 'django.core.files.storage.FileSystemStorage'
        self.defined_storage.kwargs = {
            'location': self.document_storage_kwargs['location']
        }

        self._upload_test_document()

        self.defined_storage.dotted_path = 'mayan.apps.storage.backends.compressedstorage.ZipCompressedPassthroughStorage'
        self.defined_storage.kwargs = {
            'next_storage_backend': 'django.core.files.storage.FileSystemStorage',
            'next_storage_backend_arguments': {
                'location': self.document_storage_kwargs['location']
            }
        }

        self._call_test_management_command(
            **self._get_test_management_command_keyword_arguments()
        )

    def test_storage_processor_command_forwards(self):
        self._upload_and_call()

        with open(file=self._test_document.file_latest.file.path, mode='rb') as file_object:
            self.assertEqual(
                self.mime_type_backend.get_mime_type(
                    file_object=file_object
                ), ('application/zip', 'binary')
            )

        self.assertEqual(
            self._test_document.file_latest.checksum,
            self._test_document.file_latest.checksum_update(save=False)
        )

    def test_processor_forwards_and_reverse(self):
        self._upload_and_call()

        self._call_test_management_command(
            **self._get_test_management_command_keyword_arguments(
                reverse=True
            )
        )

        self.defined_storage.dotted_path = 'django.core.files.storage.FileSystemStorage'
        self.defined_storage.kwargs = {
            'location': self.document_storage_kwargs['location']
        }

        with open(file=self._test_document.file_latest.file.path, mode='rb') as file_object:
            self.assertNotEqual(
                self.mime_type_backend.get_mime_type(
                    file_object=file_object
                ), ('application/zip', 'binary')
            )

        self.assertEqual(
            self._test_document.file_latest.checksum,
            self._test_document.file_latest.checksum_update(save=False)
        )



import logging

from django import forms
from django.utils.translation import ugettext_lazy as _

from mayan.apps.backends.forms import FormDynamicModelBackend
from mayan.apps.documents.classes import DocumentFileAction
from mayan.apps.documents.forms.document_forms import DocumentForm
from mayan.apps.documents.literals import DEFAULT_DOCUMENT_FILE_ACTION_NAME
from mayan.apps.views.widgets import DropzoneWidget

from .models import Source
from .source_backends.base import SourceBackend

logger = logging.getLogger(name=__name__)


class NewDocumentForm(DocumentForm):
    class Meta(DocumentForm.Meta):
        exclude = ('label', 'description')


class NewDocumentFileForm(forms.Form):
    comment = forms.CharField(
        help_text=_('An optional comment to explain the upload.'),
        label=_('Comment'), required=False,
        widget=forms.widgets.Textarea(
            attrs={'rows': 4}
        )
    )
    action_name = forms.ChoiceField(
        label=_('Action'), help_text=_(
            'The action to take in regards to the pages of the new file '
            'being uploaded.'
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['action_name'].choices = DocumentFileAction.get_choices()
        self.fields['action_name'].initial = DEFAULT_DOCUMENT_FILE_ACTION_NAME


class UploadBaseForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.source = kwargs.pop('source')

        super().__init__(*args, **kwargs)


class SourceBackendSelectionForm(forms.Form):
    backend = forms.ChoiceField(
        choices=(), help_text=_(
            'The backend used to create the new source.'
        ), label=_('Backend')
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['backend'].choices = SourceBackend.get_choices()


class SourceBackendSetupDynamicForm(FormDynamicModelBackend):
    class Meta:
        fields = ('label', 'enabled', 'backend_data')
        model = Source


class WebFormUploadFormHTML5(UploadBaseForm):
    file = forms.FileField(
        label=_('File'), widget=forms.widgets.FileInput()
    )

    dropzone = forms.CharField(
        label='', required=False, widget=DropzoneWidget
    )



import importlib
import logging
from pathlib import Path
import shutil

from django.core.files.base import ContentFile

from mayan.apps.common.tests.literals import (
    TEST_COMPRESSED_FILE_CONTENTS, TEST_FILE_CONTENTS_1, TEST_FILE3_PATH,
    TEST_FILENAME1, TEST_FILENAME3
)
from mayan.apps.documents.literals import STORAGE_NAME_DOCUMENT_FILES
from mayan.apps.permissions.tests.mixins import PermissionTestMixin
from mayan.apps.smart_settings.classes import SettingNamespace

from ..classes import DefinedStorage
from ..compressed_files import Archive
from ..models import DownloadFile, SharedUploadedFile
from ..utils import mkdtemp

from .literals import (
    TEST_DOWNLOAD_FILE_CONTENT_FILE_NAME, TEST_SHARED_UPLOADED_FILE_FILENAME
)


class ArchiveClassTestCaseMixin:
    archive_path = None
    cls = None
    filename = TEST_FILENAME3
    file_path = TEST_FILE3_PATH
    members_list = TEST_COMPRESSED_FILE_CONTENTS
    member_name = TEST_FILENAME1
    member_contents = TEST_FILE_CONTENTS_1

    def test_add_file(self):
        archive = self.cls()
        archive.create()
        with open(file=self.file_path, mode='rb') as file_object:
            archive.add_file(file_object=file_object, filename=self.filename)
            self.assertEqual(
                archive.members(), [self.filename]
            )

    def test_open(self):
        with open(file=self.archive_path, mode='rb') as file_object:
            archive = Archive.open(file_object=file_object)
            self.assertTrue(
                isinstance(archive, self.cls)
            )

    def test_members(self):
        with open(file=self.archive_path, mode='rb') as file_object:
            archive = Archive.open(file_object=file_object)
            self.assertEqual(
                archive.members(), self.members_list
            )

    def test_member_contents(self):
        with open(file=self.archive_path, mode='rb') as file_object:
            archive = Archive.open(file_object=file_object)
            self.assertEqual(
                archive.member_contents(filename=self.member_name),
                self.member_contents
            )

    def test_open_member(self):
        with open(file=self.archive_path, mode='rb') as file_object:
            archive = Archive.open(file_object=file_object)
            file_object = archive.open_member(filename=self.member_name)
            self.assertEqual(
                file_object.read(), self.member_contents
            )


class DownloadFileAPIViewTestMixin:
    def _request_test_download_file_delete_api_view(self):
        return self.delete(
            viewname='rest_api:download_file-detail',
            kwargs={'download_file_id': self._test_download_file.pk}
        )

    def _request_test_download_file_detail_api_view(self):
        return self.get(
            viewname='rest_api:download_file-detail',
            kwargs={'download_file_id': self._test_download_file.pk}
        )

    def _request_test_download_file_download_api_view(self):
        return self.get(
            viewname='rest_api:download_file-download',
            kwargs={'download_file_id': self._test_download_file.pk}
        )

    def _request_test_download_file_list_api_view(self):
        return self.get(viewname='rest_api:download_file-list')


class DownloadFileTestMixin(PermissionTestMixin):
    def setUp(self):
        super().setUp()
        self._test_download_file_list = []

    def _create_test_download_file(self, content=None, user=None):
        file_content = None

        test_download_file_count = len(
            self._test_download_file_list
        )

        test_download_file_filename = '{}_{}'.format(
            TEST_DOWNLOAD_FILE_CONTENT_FILE_NAME,
            test_download_file_count
        )

        if content:
            file_content = ContentFile(
                content=content, name=test_download_file_filename
            )

        self._test_download_file = DownloadFile.objects.create(
            file=file_content, user=user or self._test_case_user
        )

        self._test_download_file_list.append(self._test_download_file)


class DownloadFileViewTestMixin:
    def _request_test_download_file_delete_view(self):
        return self.post(
            viewname='storage:download_file_delete', kwargs={
                'download_file_id': self._test_download_file.pk
            }
        )

    def _request_test_download_file_download_view(self):
        return self.get(
            viewname='storage:download_file_download', kwargs={
                'download_file_id': self._test_download_file.pk
            }
        )

    def _request_test_download_file_list_view(self):
        return self.get(viewname='storage:download_file_list')


class StorageProcessorTestMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.defined_storage = DefinedStorage.get(
            name=STORAGE_NAME_DOCUMENT_FILES
        )
        cls.document_storage_kwargs = cls.defined_storage.kwargs

    def setUp(self):
        super().setUp()
        self.temporary_directory = mkdtemp()
        self.path_temporary_directory = Path(self.temporary_directory)
        self.path_test_file = self.path_temporary_directory / 'test_file.txt'

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(path=self.temporary_directory, ignore_errors=True)
        self.defined_storage.kwargs = self.document_storage_kwargs


class SharedUploadedFileTestMixin:
    def setUp(self):
        super().setUp()
        self._test_shared_uploaded_file_list = []

    def _create_test_shared_uploaded_file(self, content=None):
        file_content = None

        test_shared_uploaded_file_count = len(
            self._test_shared_uploaded_file_list
        )
        test_shared_uploaded_file_filename = '{}_{}'.format(
            TEST_SHARED_UPLOADED_FILE_FILENAME,
            test_shared_uploaded_file_count
        )

        if content:
            file_content = ContentFile(
                content=content, name=test_shared_uploaded_file_filename
            )

        self._test_shared_uploaded_file = SharedUploadedFile.objects.create(
            file=file_content, filename=test_shared_uploaded_file_filename
        )
        self._test_shared_uploaded_file_list.append(
            self._test_shared_uploaded_file
        )


class StorageSettingTestMixin:
    def _test_storage_setting_with_invalid_value(
        self, setting, storage_module, storage_name
    ):
        old_value = setting.value
        SettingNamespace.invalidate_cache_all()

        self._set_environment_variable(
            name='MAYAN_{}'.format(
                setting.global_name
            ), value='invalid_value'
        )
        self.test_case_silenced_logger_new_level = logging.FATAL + 10
        self._silence_logger(name='mayan.apps.storage.classes')

        with self.assertRaises(expected_exception=TypeError) as assertion:
            importlib.reload(storage_module)
            DefinedStorage.get(
                name=storage_name
            ).get_storage_instance()

        setting.set(value=old_value)
        importlib.reload(storage_module)

        return assertion



from django.utils.module_loading import import_string


def factory_method_periodic_task_save(super_save):
    def method_save(self, *args, **kwargs):
        import_string(dotted_path=self.task)
        return super_save(self, *args, **kwargs)
    return method_save



TEST_CABINET_LABEL = 'test cabinet label'
TEST_CABINET_LABEL_EDITED = 'test cabinet edited label'
TEST_CABINET_CHILD_LABEL = 'test cabinet child label'

TEST_INDEX_NODE_TEMPLATE = '{{ document.cabinets.first }}'



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0005_rename_filemetadataentry_field')
    ]

    operations = [
        migrations.RenameModel(
            old_name='DocumentVersionDriverEntry',
            new_name='DocumentFileDriverEntry'
        )
    ]



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0023_auto_20191213_0044')
    ]

    operations = [
        migrations.AlterField(
            model_name='intervalbasemodel',
            name='document_type',
            field=models.ForeignKey(
                help_text='Assign a document type to documents uploaded '
                'from this source.',
                on_delete=django.db.models.deletion.CASCADE,
                related_name='interval_sources', to='documents.DocumentType',
                verbose_name='Document type'
            )
        )
    ]



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..permissions import permission_document_file_view

from .mixins.document_mixins import DocumentTestMixin
from .mixins.document_file_mixins import DocumentFilePageAPIViewTestMixin


class DocumentFilePageAPIViewTestCase(
    DocumentFilePageAPIViewTestMixin, DocumentTestMixin, BaseAPITestCase
):
    def test_document_file_page_detail_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_detail_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_file,
            permission=permission_document_file_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_detail_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_file,
            permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_image_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_image_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_image_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_image_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_image_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_image_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_list_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_file,
            permission=permission_document_file_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_document_file_page.id
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_list_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_file,
            permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.document_states.permissions import permission_workflow_template_edit
from mayan.apps.document_states.tests.mixins.workflow_template_state_action_mixins import (
    WorkflowTemplateStateActionTestMixin,
    WorkflowTemplateStateActionViewTestMixin
)
from mayan.apps.testing.tests.base import BaseTestCase, GenericViewTestCase

from ..events import (
    event_cabinet_document_added, event_cabinet_document_removed
)
from ..models import Cabinet
from ..workflow_actions import CabinetAddAction, CabinetRemoveAction

from .mixins import CabinetTestMixin


class CabinetWorkflowActionTestCase(
    CabinetTestMixin, WorkflowTemplateStateActionTestMixin,
    BaseTestCase
):
    auto_create_test_cabinet = True

    def test_cabinet_document_add_action(self):
        self._clear_events()

        self._execute_workflow_template_state_action(
            klass=CabinetAddAction, kwargs={
                'cabinets': Cabinet.objects.all()
            }
        )
        self.assertTrue(
            self._test_document in self._test_cabinet.documents.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_cabinet)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_cabinet_document_added.id)

    def test_cabinet_document_remove_action(self):
        self._test_cabinet.document_add(
            document=self._test_document, user=self._test_case_user
        )

        self._clear_events()

        self._execute_workflow_template_state_action(
            klass=CabinetRemoveAction, kwargs={
                'cabinets': Cabinet.objects.all()
            }
        )
        self.assertFalse(
            self._test_document in self._test_cabinet.documents.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_cabinet)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_cabinet_document_removed.id)


class CabinetWorkflowActionViewTestCase(
    CabinetTestMixin, WorkflowTemplateStateActionViewTestMixin,
    GenericViewTestCase
):
    def test_cabinet_add_action_create_get_view(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        response = self._request_test_workflow_template_state_action_create_get_view(
            backend_path='mayan.apps.cabinets.workflow_actions.CabinetAddAction'
        )
        self.assertEqual(response.status_code, 200)

    def test_cabinet_remove_action_create_get_view(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self._create_test_cabinet()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        response = self._request_test_workflow_template_state_action_create_get_view(
            backend_path='mayan.apps.cabinets.workflow_actions.CabinetRemoveAction'
        )
        self.assertEqual(response.status_code, 200)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import DEFAULT_EVENTS_DISABLE_ASYNCHRONOUS_MODE

namespace = SettingNamespace(label=_('Events'), name='events')

setting_disable_asynchronous_mode = namespace.add_setting(
    default=DEFAULT_EVENTS_DISABLE_ASYNCHRONOUS_MODE,
    global_name='EVENTS_DISABLE_ASYNCHRONOUS_MODE',
    help_text=_(
        'Disables asynchronous events mode. All events will be committed '
        'in the same process that triggers them. This was the behavior '
        'prior to version 4.5.'
    )
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_c, worker_d

queue_signatures = CeleryQueue(
    label=_('Signatures'), name='signatures', worker=worker_c
)
queue_signatures_slow = CeleryQueue(
    label=_('Signatures slow'), name='signatures_slow', worker=worker_d
)

queue_signatures.add_task_type(
    dotted_path='mayan.apps.document_signatures.tasks.task_verify_key_signatures',
    label=_('Verify key signatures')
)
queue_signatures.add_task_type(
    dotted_path='mayan.apps.document_signatures.tasks.task_unverify_key_signatures',
    label=_('Unverify key signatures')
)
queue_signatures.add_task_type(
    dotted_path='mayan.apps.document_signatures.tasks.task_verify_document_file',
    label=_('Verify document file')
)

queue_signatures_slow.add_task_type(
    dotted_path='mayan.apps.document_signatures.tasks.task_verify_missing_embedded_signature',
    label=_('Verify missing embedded signature')
)
queue_signatures_slow.add_task_type(
    dotted_path='mayan.apps.document_signatures.tasks.task_refresh_signature_information',
    label=_('Refresh existing signature information')
)



import logging

from django import forms
from django.utils.translation import ugettext_lazy as _

from mayan.apps.sources.forms import UploadBaseForm

logger = logging.getLogger(name=__name__)


class StoredFileUploadForm(UploadBaseForm):
    """
    Form that show all the files in the source specified by the
    StoredFolderFile class passed as 'cls' argument.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            self.fields['stored_source_file_id'].choices = [
                (
                    stored_source_file.encoded_filename, str(
                        stored_source_file
                    )
                ) for stored_source_file in self.source.get_backend_instance().get_stored_file_list()
            ]
        except Exception as exception:
            logger.error('exception: %s', exception)

    stored_source_file_id = forms.ChoiceField(
        label=_('File'), widget=forms.widgets.Select(
            attrs={'class': 'select2'}
        )
    )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('metadata', '0010_auto_20180823_2353'),
    ]

    operations = [
        migrations.AlterField(
            model_name='documentmetadata',
            name='value',
            field=models.CharField(
                blank=True, db_index=True,
                help_text='The actual value stored in the metadata type '
                'field for the document.', max_length=255, null=True,
                verbose_name='Value'
            ),
        ),
        migrations.AlterField(
            model_name='metadatatype',
            name='name',
            field=models.CharField(
                help_text='Name used by other apps to reference this '
                'metadata type. Do not use python reserved words, or spaces.',
                max_length=48, unique=True, verbose_name='Name'
            ),
        ),
    ]



import json

from mayan.apps.document_states.literals import WORKFLOW_ACTION_ON_ENTRY
from mayan.apps.document_states.permissions import permission_workflow_template_edit
from mayan.apps.document_states.tests.mixins.workflow_template_state_action_mixins import (
    WorkflowTemplateStateActionTestMixin,
    WorkflowTemplateStateActionViewTestMixin
)
from mayan.apps.testing.tests.base import BaseTestCase, GenericViewTestCase

from ..workflow_actions import UpdateDocumentPageOCRAction

from .literals import (
    TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED,
    TEST_UPDATE_DOCUMENT_PAGE_OCR_ACTION_DOTTED_PATH
)
from .mixins import DocumentVersionOCRTestMixin


class UpdateDocumentPageOCRActionTestCase(
    DocumentVersionOCRTestMixin, WorkflowTemplateStateActionTestMixin,
    BaseTestCase
):
    auto_create_test_document_version = True
    auto_create_test_document_version_page = True

    def test_workflow_action_document_version_page_update_action_no_page_condition_execution(self):
        self._create_test_document_version_ocr_content()

        test_document_version_page_ocr_content = self._test_document_version_page.ocr_content.content

        self._execute_workflow_template_state_action(
            klass=UpdateDocumentPageOCRAction, kwargs={
                'page_condition': '',
                'page_content': TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
            }
        )
        self._test_document_version_page.refresh_from_db()
        self.assertEqual(
            self._test_document_version_page.ocr_content.content,
            test_document_version_page_ocr_content
        )

    def test_workflow_action_document_version_page_update_action_template_page_content_execution(self):
        self._create_test_document_version_ocr_content()

        self._execute_workflow_template_state_action(
            klass=UpdateDocumentPageOCRAction, kwargs={
                'page_condition': '{% if "Documentation" in document_version_page.ocr_content.content %}True{% endif %}',
                'page_content': TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
            }
        )
        self._test_document_version_page.refresh_from_db()
        self.assertEqual(
            self._test_document_version_page.ocr_content.content,
            TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
        )

    def test_workflow_action_update_document_version_page_action_template_page_content_execution(self):
        self._create_test_document_version_ocr_content()

        test_document_version_page_ocr_content = self._test_document_version_page.ocr_content.content

        self._execute_workflow_template_state_action(
            klass=UpdateDocumentPageOCRAction, kwargs={
                'page_condition': '{% if "Documentation" in document_version_page.ocr_content.content %}True{% endif %}',
                'page_content': '{{ document_version_page.ocr_content.content }}+update'
            }
        )
        self._test_document_version_page.refresh_from_db()
        self.assertEqual(
            self._test_document_version_page.ocr_content.content,
            '{}+update'.format(test_document_version_page_ocr_content)
        )


class TransitionUpdateDocumentPageOCRActionTestCase(
    WorkflowTemplateStateActionTestMixin, BaseTestCase
):
    def setUp(self):
        super().setUp()
        self._test_document.delete()
        self._test_document.delete()

        self._upload_test_document()

    def test_workflow_action_update_document_version_page_execution(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._test_workflow_template_states[1].actions.create(
            backend_data=json.dumps(
                obj={
                    'page_condition': True,
                    'page_content': '{{ workflow_instance.document.label }}'
                }
            ), backend_path=TEST_UPDATE_DOCUMENT_PAGE_OCR_ACTION_DOTTED_PATH,
            label='', when=WORKFLOW_ACTION_ON_ENTRY
        )
        self._test_workflow_template.document_types.add(
            self._test_document_type
        )

        self._upload_test_document()

        self._test_document.workflows.first().do_transition(
            transition=self._test_workflow_template_transition
        )

        self.assertEqual(
            ''.join(
                self._test_document.ocr_content()
            ), self._test_document.label
        )


class UpdateDocumentPageOCRActionViewTestCase(
    WorkflowTemplateStateActionViewTestMixin, GenericViewTestCase
):
    def test_update_document_version_page_ocr_workflow_state_action_create_get_view_no_permission(self):
        action_count = self._test_workflow_template_state.actions.count()

        response = self._request_test_workflow_template_state_action_create_get_view(
            backend_path='mayan.apps.ocr.workflow_actions.UpdateDocumentPageOCRAction',
        )
        self.assertEqual(response.status_code, 404)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count
        )

    def test_update_document_version_page_ocr_workflow_state_action_create_get_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )
        action_count = self._test_workflow_template_state.actions.count()

        response = self._request_test_workflow_template_state_action_create_get_view(
            backend_path='mayan.apps.ocr.workflow_actions.UpdateDocumentPageOCRAction',
        )
        self.assertEqual(response.status_code, 200)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count
        )



import os

from django.apps import apps
from django.contrib.staticfiles.finders import AppDirectoriesFinder


class MayanAppDirectoriesFinder(AppDirectoriesFinder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for app_config in apps.get_app_configs():
            if getattr(app_config, 'has_static_media', False):
                if app_config.name not in self.apps:
                    self.storages[app_config.name] = self.storage_class(
                        os.path.join(app_config.path, self.source_dir)
                    )
                    self.apps.append(app_config.name)



from mayan.apps.testing.tests.base import BaseTestCase

from ..exceptions import DynamicSearchQueryError
from ..search_query_types import (
    QueryType, QueryTypeExact, QueryTypeFuzzy, QueryTypeGreaterThan,
    QueryTypeGreaterThanOrEqual, QueryTypeLessThan, QueryTypeLessThanOrEqual,
    QueryTypePartial, QueryTypeRange, QueryTypeRangeExclusive,
    QueryTypeRegularExpression
)

from .literals import TEST_QUERY_TYPE_ALIAS_INVALID


class SearchQueryTypeTestCase(BaseTestCase):
    _test_query_type_alias = None

    def _do_test_query_type_check(self, value='term1'):
        return QueryType.check_all(
            value='{alias}{value}'.format(
                alias=self._test_query_type_alias or '', value=value
            )
        )

    def test_default(self):
        self._test_query_type_alias = None
        query_type, value = self._do_test_query_type_check()

        self.assertEqual(value, 'term1')
        self.assertTrue(issubclass(query_type, QueryTypePartial))

    def test_exact(self):
        self._test_query_type_alias = QueryTypeExact.alias
        query_type, value = self._do_test_query_type_check()

        self.assertEqual(value, 'term1')
        self.assertTrue(issubclass(query_type, QueryTypeExact))

    def test_fuzzy(self):
        self._test_query_type_alias = QueryTypeFuzzy.alias
        query_type, value = self._do_test_query_type_check()

        self.assertEqual(value, 'term1')
        self.assertTrue(issubclass(query_type, QueryTypeFuzzy))

    def test_greater_than(self):
        self._test_query_type_alias = QueryTypeGreaterThan.alias
        query_type, value = self._do_test_query_type_check()

        self.assertEqual(value, 'term1')
        self.assertTrue(issubclass(query_type, QueryTypeGreaterThan))

    def test_greater_than_or_equal(self):
        self._test_query_type_alias = QueryTypeGreaterThanOrEqual.alias
        query_type, value = self._do_test_query_type_check()

        self.assertEqual(value, 'term1')
        self.assertTrue(issubclass(query_type, QueryTypeGreaterThanOrEqual))

    def test_invalid_with_default(self):
        self._test_query_type_alias = TEST_QUERY_TYPE_ALIAS_INVALID

        query_type, value = self._do_test_query_type_check()

        self.assertEqual(
            value, '{}term1'.format(TEST_QUERY_TYPE_ALIAS_INVALID)
        )
        self.assertTrue(issubclass(query_type, QueryTypePartial))

    def test_invalid_with_default_unset(self):
        default_query_class = QueryType._default_klass
        QueryType._default_klass = None

        self._test_query_type_alias = TEST_QUERY_TYPE_ALIAS_INVALID

        with self.assertRaises(expected_exception=DynamicSearchQueryError):
            query_type, value = self._do_test_query_type_check()

        QueryType._default_klass = default_query_class

    def test_less_than(self):
        self._test_query_type_alias = QueryTypeLessThan.alias
        query_type, value = self._do_test_query_type_check()

        self.assertEqual(value, 'term1')
        self.assertTrue(issubclass(query_type, QueryTypeLessThan))

    def test_less_than_or_equal(self):
        self._test_query_type_alias = QueryTypeLessThanOrEqual.alias
        query_type, value = self._do_test_query_type_check()

        self.assertEqual(value, 'term1')
        self.assertTrue(issubclass(query_type, QueryTypeLessThanOrEqual))

    def test_partial(self):
        self._test_query_type_alias = QueryTypePartial.alias
        query_type, value = self._do_test_query_type_check()

        self.assertEqual(value, 'term1')
        self.assertTrue(issubclass(query_type, QueryTypePartial))

    def test_range(self):
        self._test_query_type_alias = QueryTypeRange.alias
        query_type, value = self._do_test_query_type_check(
            value='term1..term2'
        )

        self.assertEqual(value, 'term1..term2')
        self.assertTrue(issubclass(query_type, QueryTypeRange))

    def test_range_exclusive(self):
        self._test_query_type_alias = QueryTypeRangeExclusive.alias
        query_type, value = self._do_test_query_type_check(
            value='term1..term2'
        )

        self.assertEqual(value, 'term1..term2')
        self.assertTrue(issubclass(query_type, QueryTypeRangeExclusive))

    def test_range_regular_expression(self):
        self._test_query_type_alias = QueryTypeRegularExpression.alias
        query_type, value = self._do_test_query_type_check()

        self.assertEqual(value, 'term1')
        self.assertTrue(issubclass(query_type, QueryTypeRegularExpression))



from datetime import timedelta
import logging

from django.db import models
from django.utils.timezone import now

from .settings import (
    setting_download_file_expiration_interval,
    setting_shared_uploaded_file_expiration_interval
)

logger = logging.getLogger(name=__name__)


class DownloadFileManager(models.Manager):
    def get_stale_queryset(self):
        return self.filter(
            datetime__lt=now() - timedelta(
                seconds=setting_download_file_expiration_interval.value
            )
        )

    def stale_delete(self):
        for stale in self.get_stale_queryset():
            try:
                stale.delete()
            except Exception as exception:
                logger.error(
                    'Unable to delete stale download file ID: %d; %s',
                    stale.pk, exception
                )


class SharedUploadedFileManager(models.Manager):
    def get_stale_queryset(self):
        return self.filter(
            datetime__lt=now() - timedelta(
                seconds=setting_shared_uploaded_file_expiration_interval.value
            )
        )

    def stale_delete(self):
        for stale in self.get_stale_queryset():
            try:
                stale.delete()
            except Exception as exception:
                logger.error(
                    'Unable to delete stale shared uploaded file ID: %d; %s',
                    stale.pk, exception
                )



from django.contrib.auth import get_user_model

from mayan.apps.common.tests.mixins import ManagementCommandTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..events import event_otp_disabled
from ..literals import (
    COMMAND_NAME_AUTHENTICATION_OTP_DISABLE,
    COMMAND_NAME_AUTHENTICATION_OTP_INITIALIZE,
    COMMAND_NAME_AUTHENTICATION_OTP_STATUS
)
from ..models import UserOTPData

from .mixins import AuthenticationOTPTestMixin


class AuthenticationOTPDisableManagementCommandTestCase(
    AuthenticationOTPTestMixin, ManagementCommandTestMixin, BaseTestCase
):
    _test_management_command_name = COMMAND_NAME_AUTHENTICATION_OTP_DISABLE
    create_test_case_super_user = True

    def test_command(self):
        self._enable_test_otp()

        self._clear_events()

        self._call_test_management_command(
            self._test_case_super_user.username
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_super_user)
        self.assertEqual(events[0].target, self._test_case_super_user)
        self.assertEqual(events[0].verb, event_otp_disabled.id)


class AuthenticationOTPInitializeManagementCommandTestCase(
    AuthenticationOTPTestMixin, ManagementCommandTestMixin, BaseTestCase
):
    _test_management_command_name = COMMAND_NAME_AUTHENTICATION_OTP_INITIALIZE
    create_test_case_super_user = True

    def test_command(self):
        test_user_count = get_user_model().objects.count()

        UserOTPData.objects.all().delete()

        self._clear_events()

        self._call_test_management_command()

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

        self.assertEqual(
            UserOTPData.objects.count(), test_user_count
        )


class AuthenticationOTPStatusManagementCommandTestCase(
    AuthenticationOTPTestMixin, ManagementCommandTestMixin, BaseTestCase
):
    _test_management_command_name = COMMAND_NAME_AUTHENTICATION_OTP_STATUS
    create_test_case_super_user = True

    def test_command_with_otp_disabled(self):
        self._clear_events()

        stdout, stderr = self._call_test_management_command(
            self._test_case_super_user.username
        )
        self.assertTrue('disabled' in stdout)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_command_with_otp_enabled(self):
        self._enable_test_otp()

        self._clear_events()

        stdout, stderr = self._call_test_management_command(
            self._test_case_super_user.username
        )
        self.assertTrue('enabled' in stdout)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.backends.class_mixins import DynamicFormBackendMixin
from mayan.apps.backends.classes import ModelBaseBackend

from ..exceptions import SourceActionExceptionUnknown

from .mixins import SourceBackendMixinSourceMetadata


class SourceBackendBase:
    """
    Base class for the source backends.
    """
    action_class_list = None

    def callback_post_document_create(self, document, **kwargs):
        """
        Callback to execute when a document is created.
        """
        return

    def callback_post_document_file_create(self, document_file, **kwargs):
        """
        Callback to execute when a document file is first created.
        """
        return

    def callback_post_document_file_upload(self, document_file, **kwargs):
        """
        Callback to execute when a document file is fully uploaded.
        """
        return

    def clean(self):
        """
        Optional method to validate backend data before saving.
        """

    def create(self):
        """
        Called after the source model's .save() method for new
        instances.
        """

    def delete(self):
        """
        Called before the source model's .delete() method.
        """

    def get_action(self, name):
        for entry in self.get_action_list():
            if entry.name == name:
                return entry

        raise SourceActionExceptionUnknown(
            'Unknown action `{}` for source `{}`.'.format(
                name, self.label
            )
        )

    def get_action_class_list(self):
        """
        Returns the non initialized list of action classes. This is to allow
        mixins to add their own actions to a base source backend class.
        """
        return self.action_class_list or ()

    def get_action_list(self):
        action_class_list = self.get_action_class_list() or ()

        source = self.get_model_instance()

        for action_class in action_class_list:
            yield action_class(source=source)

    def update(self):
        """
        Called after the source model's .save() method for existing
        instances.
        """


class SourceBackend(
    DynamicFormBackendMixin, ModelBaseBackend,
    SourceBackendMixinSourceMetadata, SourceBackendBase
):
    """
    Final `SourceBackend` class. Separate from `SourceBackendBase` to allow
    for `ModelBaseBackend` to initialize properly while avoiding overriding
    mixin methods like the ones in `SourceBackendMixinSourceMetadata`.
    """
    _backend_app_label = 'sources'
    _backend_model_name = 'Source'
    _loader_module_name = 'source_backends'

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = (
            (
                _('General'), {
                    'fields': ('label', 'enabled')
                }
            ),
        )

        return fieldsets

    @classmethod
    def intialize(cls):
        """
        Optional method for subclasses execute their own initialization
        code.
        """

    @classmethod
    def post_load_modules(cls):
        for source_backend in cls.get_all():
            source_backend.intialize()

    def get_allow_action_execute(self, action, action_execute_kwargs=None):
        return self.get_model_instance().enabled

    def get_upload_form_class(self, action):
        # Hidden import to avoid model that are not ready yet. This happens
        # as `DocumentForm` is imported in `forms.py`.
        from ..forms import UploadBaseForm

        return getattr(self, 'upload_form_class', UploadBaseForm)


class SourceBackendNull(SourceBackend):
    is_visible = False
    label = _('Null backend')



from django.db import models

from ...search_backends import SearchBackend
from ...search_models import SearchModel
from ...search_query_types import QueryType, QueryTypeExact

from ..backends import TestSearchBackendProxy
from ..literals import (
    TEST_OBJECT_BOOLEAN_VALUE, TEST_OBJECT_CHAR_VALUE,
    TEST_OBJECT_EMAIL_VALUE, TEST_OBJECT_INTEGER_VALUE,
    TEST_OBJECT_POSITIVE_INTEGER_VALUE, TEST_OBJECT_TEXT_VALUE,
    TEST_OBJECT_UUID_VALUE
)


class TestSearchObjectHierarchyTestMixin:
    auto_test_search_objects_create = True

    def _create_test_models(self):
        self.TestModelAttribute = self._create_test_model(
            fields={
                'label': models.CharField(
                    max_length=32
                )
            }, model_name='TestModelAttribute'
        )
        self.TestModelGrandParent = self._create_test_model(
            fields={
                'label': models.CharField(
                    max_length=32
                )
            }, model_name='TestModelGrandParent'
        )
        self.TestModelParent = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelGrandParent',
                ),
                'label': models.CharField(
                    max_length=32
                )
            }, model_name='TestModelParent'
        )
        self.TestModelGrandChild = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelParent',
                ),
                'attributes': models.ManyToManyField(
                    related_name='children', to='TestModelAttribute',
                ),
                'label': models.CharField(
                    max_length=32
                )
            }, model_name='TestModelGrandChild'
        )
        self.TestModelGrandChildProxy = self._create_test_model(
            base_class=self.TestModelGrandChild,
            model_name='TestModelGrandChildProxy',
            options={
                'proxy': True
            }
        )

    def _create_test_search_objects(self):
        self._test_object_grandparent = self.TestModelGrandParent.objects.create(
            label='grandparent'
        )
        self._test_object_parent = self.TestModelParent.objects.create(
            parent=self._test_object_grandparent,
            label='parent'
        )
        self._test_object_grandchild = self.TestModelGrandChild.objects.create(
            parent=self._test_object_parent,
            label='grandchild_0'
        )
        self._test_object_grandchildren = [self._test_object_grandchild]
        self._test_object_grandchild = self.TestModelGrandChild.objects.create(
            parent=self._test_object_parent,
            label='grandchild_1'
        )
        self._test_object_grandchildren.append(self._test_object_grandchild)
        self._test_object_grandchild_proxy = self.TestModelGrandChildProxy.objects.get(
            pk=self._test_object_grandchild.pk
        )

        self._test_object_attribute = self.TestModelAttribute.objects.create(
            label='attribute'
        )
        self._test_object_grandchild.attributes.add(
            self._test_object_attribute
        )

        self._test_object = self._test_object_attribute

    def _setup_test_model_search(self):
        self._test_search_grandparent = SearchModel(
            app_label=self.TestModelGrandParent._meta.app_label,
            model_name=self.TestModelGrandParent._meta.model_name
        )

        self._test_search_grandparent.add_model_field(field='label')
        self._test_search_grandparent.add_model_field(
            field='children__label'
        )
        self._test_search_grandparent.add_model_field(
            field='children__children__label'
        )
        self._test_search_grandparent.add_model_field(
            field='children__children__attributes__label'
        )

        self._test_search_grandchild = SearchModel(
            app_label=self.TestModelGrandChild._meta.app_label,
            model_name=self.TestModelGrandChild._meta.model_name
        )
        self._test_search_grandchild.add_model_field(
            field='label'
        )
        self._test_search_grandchild.add_model_field(
            field='attributes__label'
        )
        self._test_search_grandchild.add_proxy_model(
            app_label=self.TestModelAttribute._meta.app_label,
            model_name='TestModelGrandChildProxy'
        )

        self._test_search_attribute = SearchModel(
            app_label=self.TestModelAttribute._meta.app_label,
            model_name=self.TestModelAttribute._meta.model_name
        )
        self._test_search_attribute.add_model_field(
            field='label'
        )
        self._test_search_attribute.add_model_field(
            field='children__label'
        )

        self._test_search_model = self._test_search_attribute


class TestSearchObjectSimpleTestMixin:
    auto_test_search_objects_create = True
    _test_object_integer_set = True

    def _create_test_models(self):
        self.TestModel = self._create_test_model(
            fields={
                'boolean': models.BooleanField(blank=True, null=True),
                'char': models.CharField(
                    blank=True, max_length=32, null=True
                ),
                'datetime': models.DateTimeField(
                    auto_now_add=True, blank=True, null=True
                ),
                'email': models.EmailField(blank=True, null=True),
                'integer': models.IntegerField(blank=True, null=True),
                'positiveinteger': models.PositiveIntegerField(
                    blank=True, null=True
                ),
                'text': models.TextField(blank=True, null=True),
                'uuid': models.UUIDField(blank=True, null=True)
            },
            model_name='TestModel'
        )

    def _create_test_search_objects(self):
        kwargs = {
            'boolean': TEST_OBJECT_BOOLEAN_VALUE,
            'char': TEST_OBJECT_CHAR_VALUE,
            'email': TEST_OBJECT_EMAIL_VALUE,
            'positiveinteger': TEST_OBJECT_POSITIVE_INTEGER_VALUE,
            'text': TEST_OBJECT_TEXT_VALUE,
            'uuid': TEST_OBJECT_UUID_VALUE
        }

        if self._test_object_integer_set:
            kwargs['integer'] = TEST_OBJECT_INTEGER_VALUE

        self._test_object = self.TestModel.objects.create(**kwargs)

    def _setup_test_model_search(self):
        self._test_search_model = SearchModel(
            app_label=self.TestModel._meta.app_label,
            model_name=self.TestModel._meta.model_name
        )

        self._test_search_model.add_model_field(field='boolean')
        self._test_search_model.add_model_field(field='char')
        self._test_search_model.add_model_field(field='datetime')
        self._test_search_model.add_model_field(field='email')
        self._test_search_model.add_model_field(field='integer')
        self._test_search_model.add_model_field(field='positiveinteger')
        self._test_search_model.add_model_field(field='text')
        self._test_search_model.add_model_field(field='uuid')


class SearchTestMixin:
    auto_test_search_backend_initialize = True
    auto_test_search_objects_create = False

    def _deindex_instance(self, instance):
        self._test_search_backend.deindex_instance(instance=instance)

    def _do_test_search(self, query):
        terms = str(
            tuple(
                query.values()
            )[0]
        ).strip()

        self.assertTrue(terms is not None)
        self.assertTrue(terms != '')

        return self._test_search_backend.search(
            search_model=self._test_search_model, query=query,
            user=self._test_case_user
        )

    def _index_instance(self, instance):
        self._test_search_backend.index_instance(instance=instance)

    def _setup_test_model_search(self):
        """
        This method allows tests to add model search configurations and
        not have to import and initialize the SearchBackend.
        """

    def setUp(self):
        self._existing_search_models = SearchModel._registry.copy()
        super().setUp()

        self._default_query_type_class = QueryType.get_default()
        QueryType.set_default(klass=QueryTypeExact)

        TestSearchBackendProxy._test_class = self.__class__
        self._test_search_backend = SearchBackend.get_instance()

        # Monkeypatch the search class so that the test behavior is only
        # enabled when called from a search test.
        self._setup_test_model_search()

        SearchBackend._enable()

        if self.auto_test_search_backend_initialize:
            self._test_search_backend._initialize()

        if self.auto_test_search_objects_create:
            self._create_test_search_objects()

        self._silence_logger(
            name='mayan.apps.dynamic_search.search_query_types'
        )

    def tearDown(self):
        self._test_search_backend.tear_down()
        self._test_search_backend.test_mode_stop()

        SearchBackend._disable()

        super().tearDown()
        QueryType.set_default(klass=self._default_query_type_class)
        SearchModel._registry = self._existing_search_models



from django.shortcuts import resolve_url

from mozilla_django_oidc.views import OIDCAuthenticationCallbackView


class MayanOIDCAuthenticationCallbackView(OIDCAuthenticationCallbackView):
    @property
    def success_url(self):
        # Pull the next url from the session or settings--we don't need to
        # sanitize here because it should already have been sanitized.
        next_url = self.request.session.get('oidc_login_next', None)
        return next_url or resolve_url(self.get_settings('LOGIN_REDIRECT_URL', '/'))



from rest_framework import status

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_web_link_navigated
from ..models import ResolvedWebLink
from ..permissions import permission_web_link_instance_view

from .mixins import ResolvedWebLinkAPIViewTestMixin


class ResolvedWebLinkAPIViewTestCase(
    DocumentTestMixin, ResolvedWebLinkAPIViewTestMixin, BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_web_link(add_test_document_type=True)

    def test_resolved_web_link_detail_api_view_no_permission(self):
        self._clear_events()

        response = self._request_resolved_web_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_detail_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )

        self._clear_events()

        response = self._request_resolved_web_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_detail_api_view_with_web_link_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        self._clear_events()

        response = self._request_resolved_web_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_detail_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )

        self._clear_events()

        response = self._request_resolved_web_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self._test_web_link.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_resolved_web_link_detail_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_resolved_web_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_resolved_web_link_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_list_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )

        self._clear_events()

        response = self._request_resolved_web_link_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_list_api_view_with_web_link_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        self._clear_events()

        response = self._request_resolved_web_link_list_api_view()

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )

        self._clear_events()

        response = self._request_resolved_web_link_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['id'], self._test_web_link.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_resolved_web_link_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_resolved_web_link_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_navigate_api_view_no_permission(self):
        self._clear_events()

        response = self._request_resolved_web_link_navigate_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_navigate_api_view_with_web_link_access(self):
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        self._clear_events()

        response = self._request_resolved_web_link_navigate_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_navigate_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )

        self._clear_events()

        response = self._request_resolved_web_link_navigate_api_view()

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_resolved_web_link_navigate_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        self._clear_events()

        response = self._request_resolved_web_link_navigate_api_view()
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(
            response.url, ResolvedWebLink.objects.get(
                pk=self._test_web_link.pk
            ).get_redirect_url_for(document=self._test_document)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_web_link)
        self.assertEqual(events[0].verb, event_web_link_navigated.id)

    def test_trashed_document_resolved_web_link_navigate_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_web_link_instance_view
        )
        self.grant_access(
            obj=self._test_web_link,
            permission=permission_web_link_instance_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_resolved_web_link_navigate_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging

from rest_framework.filters import BaseFilterBackend

from .mixins import QuerysetSearchModelMixin
from .view_mixins import QuerysetSearchFilterMixin

logger = logging.getLogger(name=__name__)


class RESTAPISearchFilter(
    QuerysetSearchFilterMixin, QuerysetSearchModelMixin, BaseFilterBackend
):
    def filter_queryset(self, request, queryset, view):
        self.search_disable_list_filtering = getattr(
            view, 'search_disable_list_filtering', False
        )

        return self.get_filtered_queryset(
            queryset=queryset, request=request, view=view
        )



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0006_auto_20170803_0651'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workflowinstancelogentry',
            name='user',
            field=models.ForeignKey(
                blank=True, null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL, verbose_name='User'
            ),
        ),
        migrations.AlterField(
            model_name='workflowtransitiontriggerevent',
            name='event_type',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='events.EventType', verbose_name='Event type'
            ),
        ),
        migrations.AlterField(
            model_name='workflowtransitiontriggerevent',
            name='transition',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='trigger_events',
                to='document_states.WorkflowTransition',
                verbose_name='Transition'
            ),
        ),
    ]

    run_before = [
        # In this migration EventType becomes StoredEventType. Therefore this
        # migration needs to run before the source is renamed.
        ('events', '0004_auto_20170731_0423')
    ]



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0073_auto_20201130_0341'),
    ]

    operations = [
        migrations.AlterField(
            model_name='duplicateddocumentold',
            name='document',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='duplicates_old', to='documents.Document',
                verbose_name='Document'
            ),
        ),
    ]



from mayan.apps.credentials.tests.mixins import StoredCredentialPasswordUsernameTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from .mixins import SourceTestMixinStoredFile


class SourceBackendStorageTestCase(SourceTestMixinStoredFile, BaseTestCase):
    _test_source_create_auto = False

    def test_storage_arguments_template_default(self):
        TEST_SOURCE_STORAGE_BACKEND_ARGUMENTS_TEXT = '{\'location\': \'\'}'
        TEST_SOURCE_STORAGE_BACKEND_ARGUMENTS_OBJECT = {'location': ''}

        self._test_source_create(
            extra_data={
                'storage_backend_arguments': TEST_SOURCE_STORAGE_BACKEND_ARGUMENTS_TEXT
            }
        )

        backend_instance = self._test_source.get_backend_instance()

        storage_backend_arguments = backend_instance.get_storage_backend_arguments()

        self.assertEqual(
            storage_backend_arguments,
            TEST_SOURCE_STORAGE_BACKEND_ARGUMENTS_OBJECT
        )


class SourceBackendStorageCredentialTestCase(
    SourceTestMixinStoredFile, StoredCredentialPasswordUsernameTestMixin,
    BaseTestCase
):
    _test_source_create_auto = False

    def test_storage_arguments_credential_template(self):
        TEST_SOURCE_STORAGE_BACKEND_ARGUMENTS_TEXT = '{\'username\': \'{{ credential.username }}\', \'password\': \'{{ credential.password }}\'}'
        TEST_SOURCE_STORAGE_BACKEND_ARGUMENTS_OBJECT = {
            'password': self._test_stored_credential._password,
            'username': self._test_stored_credential._username
        }

        self._test_source_create(
            extra_data={
                'stored_credential_id': self._test_stored_credential.pk,
                'storage_backend_arguments': TEST_SOURCE_STORAGE_BACKEND_ARGUMENTS_TEXT
            }
        )

        backend_instance = self._test_source.get_backend_instance()
        storage_backend_arguments = backend_instance.get_storage_backend_arguments()

        self.assertEqual(
            storage_backend_arguments,
            TEST_SOURCE_STORAGE_BACKEND_ARGUMENTS_OBJECT
        )



import logging

from django.apps import apps

from .settings import setting_auto_parsing

logger = logging.getLogger(name=__name__)


def handler_initialize_new_parsing_settings(sender, instance, **kwargs):
    DocumentTypeSettings = apps.get_model(
        app_label='document_parsing', model_name='DocumentTypeSettings'
    )

    if kwargs['created']:
        DocumentTypeSettings.objects.create(
            document_type=instance, auto_parsing=setting_auto_parsing.value
        )


def handler_parse_document_file(sender, instance, **kwargs):
    if instance.document.document_type.parsing_settings.auto_parsing:
        instance.submit_for_parsing()



from django.db import migrations


def code_delete_index_instance_nodes(apps, schema_editor):
    IndexInstanceNode = apps.get_model(
        app_label='document_indexing', model_name='IndexInstanceNode'
    )

    IndexInstanceNode.objects.all().delete()


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0022_indexinstance'),
    ]

    operations = [
        migrations.RunPython(
            code=code_delete_index_instance_nodes,
            reverse_code=migrations.RunPython.noop
        ),
    ]



from django.core import validators
from django.utils.translation import ugettext_lazy as _

from .utils import split_recipient_list


def validate_email_multiple(value):
    recipient_list = split_recipient_list(
        recipients=[value]
    )

    for recipient in recipient_list:
        validate_email = validators.EmailValidator(
            message=_('%(email)s is not a valid email address.') % {
                'email': recipient
            }
        )
        validate_email(value=recipient)



from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.messaging.events import event_message_created
from mayan.apps.messaging.models import Message
from mayan.apps.storage.events import event_download_file_created
from mayan.apps.storage.models import DownloadFile

from ..classes import DocumentFileCompressor
from ..events import event_document_file_downloaded

from .literals import TEST_DOCUMENT_FILE_DOWNLOAD_MESSAGE_SUBJECT


class DocumentFileCompressorClassTestCase(GenericDocumentTestCase):
    def test_document_file_download(self):
        self._create_test_user()

        download_file_count = DownloadFile.objects.count()
        message_count = Message.objects.count()

        self._clear_events()

        self._test_user.locale_profile.language = 'es'

        document_file_compressor = DocumentFileCompressor(
            queryset=DocumentFile.valid.all()
        )
        document_file_compressor.compress_to_download_file(
            user=self._test_user
        )

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count + 1
        )

        self.assertEqual(
            Message.objects.count(), message_count + 1
        )

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        self.assertEqual(
            test_message.subject, TEST_DOCUMENT_FILE_DOWNLOAD_MESSAGE_SUBJECT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_user)
        self.assertEqual(events[0].target, test_download_file)
        self.assertEqual(events[0].verb, event_download_file_created.id)

        self.assertEqual(events[1].action_object, test_download_file)
        self.assertEqual(events[1].actor, self._test_user)
        self.assertEqual(events[1].target, self._test_document_file)
        self.assertEqual(events[1].verb, event_document_file_downloaded.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, test_message)
        self.assertEqual(events[2].target, test_message)
        self.assertEqual(events[2].verb, event_message_created.id)

    def test_document_file_empty_content_type_download(self):
        self._create_test_user()

        download_file_count = DownloadFile.objects.count()
        message_count = Message.objects.count()
        self._test_document_file.delete()

        self._clear_events()

        self._test_user.locale_profile.language = 'es'

        document_file_compressor = DocumentFileCompressor(
            queryset=DocumentFile.valid.all()
        )
        document_file_compressor.compress_to_download_file(
            user=self._test_user
        )

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )

        self.assertEqual(
            Message.objects.count(), message_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.django_gpg.tests.mixins import KeyTestMixin
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.literals import TEST_FILE_SMALL_PATH

from ..models import DetachedSignature, EmbeddedSignature
from ..permissions import permission_document_file_signature_verify

from .literals import (
    TEST_SIGNED_DOCUMENT_COUNT, TEST_SIGNED_DOCUMENT_PATH,
    TEST_UNSIGNED_DOCUMENT_COUNT
)
from .mixins import DetachedSignatureTestMixin, SignatureToolsViewTestMixin


class SignatureToolsViewTestCase(
    KeyTestMixin, DetachedSignatureTestMixin, SignatureToolsViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        class DetachedSignatureProxy(DetachedSignature):
            def save(self, *args, **kwargs):
                return super(DetachedSignature, self).save(*args, **kwargs)

            class Meta:
                proxy = True

        class EmbeddedSignatureProxy(EmbeddedSignature):
            def save(self, *args, **kwargs):
                return super(EmbeddedSignature, self).save(*args, **kwargs)

            class Meta:
                proxy = True

        cls.DetachedSignatureProxy = DetachedSignatureProxy
        cls.EmbeddedSignatureProxy = EmbeddedSignatureProxy

    def test_missing_signature_verify_view_no_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        old_hooks = DocumentFile._post_save_hooks
        DocumentFile._post_save_hooks = {}

        self._test_document_path = TEST_FILE_SMALL_PATH
        for count in range(TEST_UNSIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        for count in range(TEST_SIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT + TEST_SIGNED_DOCUMENT_COUNT
        )

        DocumentFile._post_save_hooks = old_hooks

        self._clear_events()

        response = self._request_all_test_document_file_signature_verify_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT + TEST_SIGNED_DOCUMENT_COUNT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_missing_signature_verify_view_with_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        old_hooks = DocumentFile._post_save_hooks
        DocumentFile._post_save_hooks = {}

        self._test_document_path = TEST_FILE_SMALL_PATH
        for count in range(TEST_UNSIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        for count in range(TEST_SIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT + TEST_SIGNED_DOCUMENT_COUNT
        )

        DocumentFile._post_save_hooks = old_hooks

        self.grant_permission(
            permission=permission_document_file_signature_verify
        )

        self._clear_events()

        response = self._request_all_test_document_file_signature_verify_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_missing_signature_verify_view_with_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        old_hooks = DocumentFile._post_save_hooks
        DocumentFile._post_save_hooks = {}

        self._test_document_path = TEST_FILE_SMALL_PATH
        for count in range(TEST_UNSIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        for count in range(TEST_SIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT + TEST_SIGNED_DOCUMENT_COUNT
        )

        DocumentFile._post_save_hooks = old_hooks

        self.grant_permission(
            permission=permission_document_file_signature_verify
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_all_test_document_file_signature_verify_view()

        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_refresh_view_no_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature = self.DetachedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        signature = self.EmbeddedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        self._clear_events()

        response = self._request_all_test_document_file_signature_refresh_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            DetachedSignature.objects.first().date_time, None
        )
        self.assertEqual(
            EmbeddedSignature.objects.first().date_time, None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_signature_refresh_view_with_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature = self.DetachedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        signature = self.EmbeddedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        self.grant_permission(
            permission=permission_document_file_signature_verify
        )

        self._clear_events()

        response = self._request_all_test_document_file_signature_refresh_view()
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(
            DetachedSignature.objects.first().date_time, None
        )
        self.assertNotEqual(
            EmbeddedSignature.objects.first().date_time, None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_refresh_view_with_permission(self):
        # Silence converter logging.
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature = self.DetachedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        signature = self.EmbeddedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        self._test_documents[0].delete()
        self._test_documents[1].delete()

        self.grant_permission(
            permission=permission_document_file_signature_verify
        )

        self._clear_events()

        response = self._request_all_test_document_file_signature_refresh_view()
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(
            DetachedSignature.objects.first().date_time, None
        )
        self.assertNotEqual(
            EmbeddedSignature.objects.first().date_time, None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Messaging'), name='messaging'
)

permission_message_create = namespace.add_permission(
    label=_('Create messages'), name='message_create'
)
permission_message_delete = namespace.add_permission(
    label=_('Delete messages'), name='message_delete'
)
permission_message_edit = namespace.add_permission(
    label=_('Edit messages'), name='message_edit'
)
permission_message_view = namespace.add_permission(
    label=_('View messages'), name='message_view'
)



from django.conf import settings
from django.db import migrations


def code_add_user_theme_settings_to_existing_users(apps, schema_editor):
    User = apps.get_model(settings.AUTH_USER_MODEL)
    UserThemeSetting = apps.get_model(
        app_label='appearance', model_name='UserThemeSetting'
    )

    for user in User.objects.using(alias=schema_editor.connection.alias).all():
        UserThemeSetting.objects.using(
            alias=schema_editor.connection.alias
        ).create(user=user)


def code_remove_user_theme_settings_from_existing_users(apps, schema_editor):
    User = apps.get_model(settings.AUTH_USER_MODEL)
    UserThemeSetting = apps.get_model(
        app_label='appearance', model_name='UserThemeSetting'
    )

    for user in User.objects.using(alias=schema_editor.connection.alias).all():
        UserThemeSetting.objects.using(
            alias=schema_editor.connection.alias
        ).filter(user=user).delete()


class Migration(migrations.Migration):
    dependencies = [
        ('appearance', '0001_initial')
    ]

    operations = [
        migrations.RunPython(
            code=code_add_user_theme_settings_to_existing_users,
            reverse_code=code_remove_user_theme_settings_from_existing_users
        )
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0008_alter_filemetadataentry_value')
    ]

    operations = [
        migrations.AlterField(
            model_name='documenttypesettings', name='auto_process',
            field=models.BooleanField(
                default=True, help_text='Automatically queue newly '
                'created documents for processing.',
                verbose_name='Auto process'
            )
        )
    ]



from unittest import skip

from selenium.common.exceptions import NoAlertPresentException

from django.conf import settings
from django.contrib.staticfiles.testing import StaticLiveServerTestCase

from mayan.apps.user_management.permissions import (
    permission_user_edit, permission_user_view
)
from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.testing.tests.mixins import SeleniumTestMixin

from ..events import (
    event_theme_created, event_theme_edited, event_user_theme_settings_edited
)
from ..models import Theme
from ..permissions import (
    permission_theme_create, permission_theme_delete, permission_theme_edit,
    permission_theme_view
)

from .mixins import (
    ThemeTestMixin, ThemeViewTestMixin, UserThemeSettingsViewTestMixin
)


@skip(reason='Skip until a synchronous live server class is added.')
class BasePlainViewTestCase(
    SeleniumTestMixin, StaticLiveServerTestCase, GenericViewTestCase
):
    auto_add_test_view = True
    test_view_url = r'^javascript:alert\("XSS"\)/$'
    test_view_is_public = True
    test_view_template = 'javascript_view'

    def test_login_view_url_fragment_xss(self):
        # Should redirect and not display an alert
        self._open_url(
            fragment='#javascript:alert("XSS")', viewname=settings.LOGIN_URL
        )

        with self.assertRaises(expected_exception=NoAlertPresentException):
            self.webdriver.switch_to_alert()

    def test_login_view_url_redirect(self):
        self._open_url(
            fragment='#javascript:alert("XSS")', viewname=settings.LOGIN_URL
        )

        self.assertTrue(
            self.test_view_template in self.webdriver.page_source
        )


class ThemeViewTestCase(ThemeViewTestMixin, GenericViewTestCase):
    def test_theme_create_view_no_permission(self):
        theme_count = Theme.objects.count()

        self._clear_events()

        response = self._request_test_theme_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            Theme.objects.count(), theme_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_theme_create_view_with_permissions(self):
        self.grant_permission(permission=permission_theme_create)

        theme_count = Theme.objects.count()

        self._clear_events()

        response = self._request_test_theme_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Theme.objects.count(), theme_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_theme)
        self.assertEqual(events[0].verb, event_theme_created.id)

    def test_theme_delete_view_no_permission(self):
        self._create_test_theme()

        theme_count = Theme.objects.count()

        self._clear_events()

        response = self._request_test_theme_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Theme.objects.count(), theme_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_theme_delete_view_with_access(self):
        self._create_test_theme()

        self.grant_access(
            obj=self._test_theme, permission=permission_theme_delete
        )

        theme_count = Theme.objects.count()

        self._clear_events()

        response = self._request_test_theme_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Theme.objects.count(), theme_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_theme_edit_view_no_permission(self):
        self._create_test_theme()

        theme_label = self._test_theme.label

        self._clear_events()

        response = self._request_test_theme_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_theme.refresh_from_db()
        self.assertEqual(self._test_theme.label, theme_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_theme_edit_view_with_access(self):
        self._create_test_theme()

        self.grant_access(
            obj=self._test_theme, permission=permission_theme_edit
        )

        theme_label = self._test_theme.label

        self._clear_events()

        response = self._request_test_theme_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_theme.refresh_from_db()
        self.assertNotEqual(self._test_theme.label, theme_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_theme)
        self.assertEqual(events[0].verb, event_theme_edited.id)

    def test_theme_list_view_with_no_permission(self):
        self._create_test_theme()

        self._clear_events()

        response = self._request_test_theme_list_view()
        self.assertNotContains(
            response=response, text=self._test_theme.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_theme_list_view_with_access(self):
        self._create_test_theme()

        self.grant_access(
            obj=self._test_theme, permission=permission_theme_view
        )

        self._clear_events()

        response = self._request_test_theme_list_view()
        self.assertContains(
            response=response, text=self._test_theme.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class ThemeEffectViewTestCase(ThemeTestMixin, GenericViewTestCase):
    def test_normal_view_after_theme_delete(self):
        self._create_test_theme()

        self._test_case_user.theme_settings.theme = self._test_theme
        self._test_case_user.theme_settings.save()
        self._test_theme.delete()

        response = self.get(viewname='common:about_view')

        self.assertEqual(response.status_code, 200)


class CurrentUserViewTestCase(
    UserThemeSettingsViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_theme()

    def test_current_user_theme_settings_detail_view_no_permission(self):
        self._clear_events()

        response = self._request_test_current_user_theme_settings_detail_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_current_user_theme_settings_edit_view_no_permission(self):
        theme = self._test_case_user.theme_settings.theme

        self._clear_events()

        response = self._request_test_current_user_theme_settings_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_case_user.refresh_from_db()
        self.assertNotEqual(self._test_case_user.theme_settings.theme, theme)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_case_user)
        self.assertEqual(events[0].verb, event_user_theme_settings_edited.id)


class SuperUserThemeSettingsViewTestCase(
    UserThemeSettingsViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_super_user()
        self._create_test_theme()

    def test_super_user_theme_settings_detail_view_no_permission(self):
        self._clear_events()

        response = self._request_test_super_user_theme_settings_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_super_user_theme_settings_detail_view_with_access(self):
        self.grant_access(
            obj=self._test_super_user, permission=permission_user_view
        )

        self._clear_events()

        response = self._request_test_super_user_theme_settings_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_super_user_theme_settings_edit_view_no_permission(self):
        theme = self._test_super_user.theme_settings.theme

        self._clear_events()

        response = self._request_test_super_user_theme_settings_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_super_user.refresh_from_db()
        self.assertEqual(self._test_super_user.theme_settings.theme, theme)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_super_user_theme_settings_edit_view_with_access(self):
        theme = self._test_super_user.theme_settings.theme

        self.grant_access(
            obj=self._test_super_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_super_user_theme_settings_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_super_user.refresh_from_db()
        self.assertEqual(self._test_super_user.theme_settings.theme, theme)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class UserThemeSettingsViewTestCase(
    UserThemeSettingsViewTestMixin, GenericViewTestCase
):
    auto_create_test_user = True

    def setUp(self):
        super().setUp()
        self._create_test_theme()

    def test_user_theme_settings_detail_view_no_permission(self):
        self._clear_events()

        response = self._request_test_user_theme_settings_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_theme_settings_detail_view_with_access(self):
        self.grant_access(
            obj=self._test_user, permission=permission_user_view
        )

        self._clear_events()

        response = self._request_test_user_theme_settings_detail_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_theme_settings_edit_view_no_permission(self):
        theme = self._test_user.theme_settings.theme

        self._clear_events()

        response = self._request_test_user_theme_settings_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_user.refresh_from_db()
        self.assertEqual(self._test_user.theme_settings.theme, theme)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_theme_settings_edit_view_with_access(self):
        theme = self._test_user.theme_settings.theme

        self.grant_access(
            obj=self._test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_theme_settings_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_user.refresh_from_db()
        self.assertNotEqual(self._test_user.theme_settings.theme, theme)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user)
        self.assertEqual(events[0].verb, event_user_theme_settings_edited.id)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from .model_mixins import StatisticResultBusinessLogicMixin


class StatisticResult(StatisticResultBusinessLogicMixin, models.Model):
    # Translators: 'Slug' refers to the URL valid ID of the statistic
    # More info: https://docs.djangoproject.com/en/1.7/glossary/#term-slug
    slug = models.SlugField(
        unique=True, verbose_name=_('Slug')
    )
    datetime = models.DateTimeField(
        auto_now=True, verbose_name=_('Date time')
    )
    serialize_data = models.TextField(
        blank=True, verbose_name=_('Data')
    )

    class Meta:
        verbose_name = _('Statistics result')
        verbose_name_plural = _('Statistics results')

    def __str__(self):
        return self.slug



import csv
import logging

from furl import furl

from django.apps import apps
from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from actstream import action

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.menus import menu_list_facet
from mayan.apps.organizations.utils import get_organization_installation_url

from .literals import (
    DEFAULT_EVENT_LIST_EXPORT_FILENAME, EVENT_TYPE_NAMESPACE_NAME,
    EVENT_EVENTS_CLEARED_NAME, EVENT_EVENTS_EXPORTED_NAME
)
from .links import (
    link_object_event_list, link_object_event_type_user_subscription_list
)
from .permissions import (
    permission_events_clear, permission_events_export, permission_events_view
)
from .settings import setting_disable_asynchronous_mode

logger = logging.getLogger(name=__name__)


DEFAULT_ACTION_EXPORTER_FIELD_NAMES = (
    'timestamp', 'id', 'actor_content_type', 'actor_object_id', 'actor',
    'target_content_type', 'target_object_id', 'target', 'verb',
    'action_object_content_type', 'action_object_object_id', 'action_object'
)


class ActionExporter:
    def __init__(self, queryset, field_names=None):
        self.field_names = field_names or DEFAULT_ACTION_EXPORTER_FIELD_NAMES
        self.queryset = queryset

    def export(self, file_object, user=None):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        if user:
            self.queryset = AccessControlList.objects.restrict_queryset(
                queryset=self.queryset,
                permission=permission_events_export,
                user=user
            )

        writer = csv.writer(
            file_object, delimiter=',', quotechar='"',
            quoting=csv.QUOTE_MINIMAL
        )
        file_object.write(
            ','.join(
                self.field_names + ('\n',)
            )
        )

        for entry in self.queryset.iterator():
            row = [
                str(
                    getattr(entry, field_name)
                ) for field_name in self.field_names
            ]
            writer.writerow(row)

    def export_to_download_file(
        self, organization_installation_url=None, user=None
    ):
        event_type_namespace = EventTypeNamespace.get(
            name=EVENT_TYPE_NAMESPACE_NAME
        )
        event_events_exported = event_type_namespace.get_event(
            name=EVENT_EVENTS_EXPORTED_NAME
        )

        DownloadFile = apps.get_model(
            app_label='storage', model_name='DownloadFile'
        )
        Message = apps.get_model(
            app_label='messaging', model_name='Message'
        )

        download_file = DownloadFile(
            filename=DEFAULT_EVENT_LIST_EXPORT_FILENAME,
            label=_('Event list export to CSV'), user=user
        )
        download_file._event_actor = user
        download_file.save()

        with download_file.open(mode='w') as file_object:
            self.export(file_object=file_object, user=user)

        event_events_exported.commit(
            actor=user, target=download_file
        )

        if user:
            if not organization_installation_url:
                organization_installation_url = get_organization_installation_url()

            download_list_url = furl(organization_installation_url).join(
                reverse(
                    viewname='storage:download_file_list'
                )
            ).tostr()
            download_url = furl(organization_installation_url).join(
                reverse(
                    viewname='storage:download_file_download',
                    kwargs={'download_file_id': download_file.pk}
                )
            ).tostr()

            Message.objects.create(
                sender_object=download_file,
                user=user,
                subject=_('Events exported.'),
                body=_(
                    'The event list has been exported and is available '
                    'for download using the link: %(download_url)s or from '
                    'the downloads area (%(download_list_url)s).'
                ) % {
                    'download_list_url': download_list_url,
                    'download_url': download_url
                }
            )


class EventModelRegistry:
    _registry = set()

    @classmethod
    def register(
        cls, model, acl_bind_link=True, bind_events_link=True,
        bind_subscription_link=True, exclude=None, menu=None,
        register_permissions=True
    ):
        # Hidden imports.
        from actstream import registry
        from mayan.apps.acls.classes import ModelPermission

        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        StoredPermission = apps.get_model(
            app_label='permissions', model_name='StoredPermission'
        )

        event_type_namespace = EventTypeNamespace.get(
            name=EVENT_TYPE_NAMESPACE_NAME
        )
        event_events_cleared = event_type_namespace.get_event(
            name=EVENT_EVENTS_CLEARED_NAME
        )
        event_events_exported = event_type_namespace.get_event(
            name=EVENT_EVENTS_EXPORTED_NAME
        )

        if model not in cls._registry:
            cls._registry.add(model)
            # These need to happen only once.
            registry.register(model)

            menu = menu or menu_list_facet

            if bind_events_link:
                menu.bind_links(
                    exclude=exclude,
                    links=(
                        link_object_event_list,
                    ), sources=(model,)
                )

            if bind_subscription_link:
                menu.bind_links(
                    exclude=exclude,
                    links=(
                        link_object_event_type_user_subscription_list,
                    ), sources=(model,)
                )

            if register_permissions and not issubclass(model, (AccessControlList, StoredPermission)):
                ModelPermission.register(
                    exclude=exclude,
                    model=model, permissions=(
                        permission_events_clear, permission_events_export,
                        permission_events_view
                    ), bind_link=acl_bind_link
                )

                ModelEventType.register(
                    event_types=(
                        event_events_cleared, event_events_exported
                    ), model=model
                )


class EventTypeNamespace(AppsModuleLoaderMixin):
    _registry = {}
    _loader_module_name = 'events'

    @classmethod
    def all(cls):
        return sorted(
            cls._registry.values()
        )

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    def __init__(self, name, label):
        self.name = name
        self.label = label
        self.event_types = []
        self.__class__._registry[name] = self

    def __lt__(self, other):
        return self.label < other.label

    def __str__(self):
        return str(self.label)

    def add_event_type(self, name, label):
        event_type = EventType(namespace=self, name=name, label=label)
        self.event_types.append(event_type)
        return event_type

    def get_event(self, name):
        return EventType.get(
            id='{}.{}'.format(self.name, name)
        )

    def get_event_types(self):
        return EventType.sort(event_type_list=self.event_types)


class EventType:
    _registry = {}

    @staticmethod
    def sort(event_type_list):
        return sorted(
            event_type_list, key=lambda event_type: (
                event_type.namespace.label, event_type.label
            )
        )

    @classmethod
    def all(cls):
        # Return sorted permisions by namespace.name.
        return EventType.sort(
            event_type_list=cls._registry.values()
        )

    @classmethod
    def get(cls, id):
        return cls._registry[id]

    @classmethod
    def refresh(cls):
        for event_type in cls.all():
            # Invalidate cache and recreate store events while repopulating
            # cache.
            event_type.stored_event_type = None
            event_type.get_stored_event_type()

    def __init__(self, namespace, name, label):
        self.namespace = namespace
        self.name = name
        self.label = label
        self.stored_event_type = None
        self.__class__._registry[self.id] = self

    def __str__(self):
        return '{}: {}'.format(self.namespace.label, self.label)

    def _commit(self, action_object=None, actor=None, target=None):
        EventSubscription = apps.get_model(
            app_label='events', model_name='EventSubscription'
        )
        Notification = apps.get_model(
            app_label='events', model_name='Notification'
        )
        ObjectEventSubscription = apps.get_model(
            app_label='events', model_name='ObjectEventSubscription'
        )
        User = get_user_model()

        if actor is None and target is None:
            # If the actor and the target are None there is no way to
            # create a new event.
            logger.warning(
                'Attempting to commit event "%s" without an actor or a '
                'target. This is not supported.', self
            )
            return

        result = action.send(
            actor or target, actor=actor, verb=self.id,
            action_object=action_object, target=target
        )[0][1]
        # The [0][1] means: get the first and only action from the list
        # and ignore the handler.

        # Create notifications for the actions created by the event committed.

        # Gather the users subscribed globally to the event.
        user_queryset = User.objects.filter(
            id__in=EventSubscription.objects.filter(
                stored_event_type__name=result.verb
            ).values('user')
        )

        # Gather the users subscribed to the target object event.
        if result.target:
            user_queryset = user_queryset | User.objects.filter(
                id__in=ObjectEventSubscription.objects.filter(
                    content_type=result.target_content_type,
                    object_id=result.target.pk,
                    stored_event_type__name=result.verb
                ).values('user')
            )

        # Gather the users subscribed to the action object event.
        if result.action_object:
            user_queryset = user_queryset | User.objects.filter(
                id__in=ObjectEventSubscription.objects.filter(
                    content_type=result.action_object_content_type,
                    object_id=result.action_object.pk,
                    stored_event_type__name=result.verb
                ).values('user')
            )

        for user in user_queryset:
            if result.action_object:
                Notification.objects.create(action=result, user=user)
                # Don't check or add any other notification for the
                # same user-event-object.
                continue

            if result.target:
                Notification.objects.create(action=result, user=user)
                # Don't check or add any other notification for the
                # same user-event-object.
                continue

    def commit(self, action_object=None, actor=None, target=None):
        # Hidden import.
        from .tasks import task_event_commit
        # This circular import is necessary.

        task_kwargs = {'event_id': self.id}

        if setting_disable_asynchronous_mode.value:
            self._commit(
                action_object=action_object, actor=actor, target=target
            )
        else:
            if action_object:
                task_kwargs.update(
                    {
                        'action_object_app_label': action_object._meta.app_label,
                        'action_object_model_name': action_object._meta.model_name,
                        'action_object_id': action_object.pk
                    }
                )

            if actor:
                task_kwargs.update(
                    {
                        'actor_app_label': actor._meta.app_label,
                        'actor_model_name': actor._meta.model_name,
                        'actor_id': actor.pk
                    }
                )

            if target:
                task_kwargs.update(
                    {
                        'target_app_label': target._meta.app_label,
                        'target_model_name': target._meta.model_name,
                        'target_id': target.pk
                    }
                )

            task_event_commit.apply_async(kwargs=task_kwargs)

    def get_stored_event_type(self):
        if not self.stored_event_type:
            StoredEventType = apps.get_model(
                app_label='events', model_name='StoredEventType'
            )

            self.stored_event_type, created = StoredEventType.objects.get_or_create(
                name=self.id
            )

        return self.stored_event_type

    @property
    def id(self):
        return '{}.{}'.format(self.namespace.name, self.name)


class ModelEventType:
    """
    Class to allow matching a model to a specific set of events.
    """
    _inheritances = {}
    _registry = {}

    @classmethod
    def get_for_class(cls, klass):
        result = cls._registry.get(
            klass, ()
        )
        return EventType.sort(event_type_list=result)

    @classmethod
    def get_for_instance(cls, instance):
        StoredEventType = apps.get_model(
            app_label='events', model_name='StoredEventType'
        )

        events = []

        class_events = cls._registry.get(
            type(instance)
        )

        if class_events:
            events.extend(class_events)

        pks = [
            event.id for event in set(events)
        ]

        return EventType.sort(
            event_type_list=StoredEventType.objects.filter(name__in=pks)
        )

    @classmethod
    def get_inheritance(cls, model):
        return cls._inheritances[model]

    @classmethod
    def register(cls, model, event_types):
        cls._registry.setdefault(
            model, []
        )
        for event_type in event_types:
            cls._registry[model].append(event_type)

    @classmethod
    def register_inheritance(cls, model, related):
        cls._inheritances[model] = related
